from warnings import warn

import numpy as np

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.evaluation import after_stat
from .binning import (
    assign_bins,
    breaks_from_bins,
    breaks_from_binwidth,
    freedman_diaconis_bins,
)
from .stat import stat


@document
class stat_bin(stat):
    """
    Count cases in each interval

    {usage}

    Parameters
    ----------
    {common_parameters}
    binwidth : float, optional (default: None)
        The width of the bins. The default is to use bins bins that
        cover the range of the data. You should always override this
        value, exploring multiple widths to find the best to illustrate
        the stories in your data.
    bins : int, optional (default: None)
        Number of bins. Overridden by binwidth. If :py:`None`,
        a number is computed using the freedman-diaconis method.
    breaks : array-like, optional (default: None)
        Bin boundaries. This supercedes the ``binwidth``, ``bins``,
        ``center`` and ``boundary``.
    center : float, optional (default: None)
        The center of one of the bins. Note that if center is above
        or below the range of the data, things will be shifted by
        an appropriate number of widths. To center on integers, for
        example, use :py:`width=1` and :py:`center=0`, even if 0 i
        s outside the range of the data. At most one of center and
        boundary may be specified.
    boundary : float, optional (default: None)
        A boundary between two bins. As with center, things are
        shifted when boundary is outside the range of the data.
        For example, to center on integers, use :py:`width=1` and
        :py:`boundary=0.5`, even if 1 is outside the range of the
        data. At most one of center and boundary may be specified.
    closed : str, optional (default: right)
        Which edge of the bins is included, *left* or *right*.
    pad : bool, optional (default: False)
        If :py:`True`, adds empty bins at either side of x.
        This ensures that frequency polygons touch 0.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'count'    # number of points in bin
         'density'  # density of points in bin, scaled to integrate to 1
         'ncount'   # count, scaled to maximum of 1
         'ndensity' # density, scaled to maximum of 1
         'ngroup'   # number of points in group

    """
    REQUIRED_AES = {"x"}
    DEFAULT_PARAMS = {
        "geom": "histogram",
        "position": "stack",
        "na_rm": False,
        "binwidth": None,
        "bins": None,
        "breaks": None,
        "center": None,
        "boundary": None,
        "closed": "right",
        "pad": False,
    }
    DEFAULT_AES = {"y": after_stat("count"), "weight": None}
    CREATES = {"width", "count", "density", "ncount", "ndensity", "ngroup"}

    def setup_params(self, data):
        params = self.params

        if "y" in data or "y" in params:
            msg = "stat_bin() must not be used with a y aesthetic."
            raise PlotnineError(msg)

        if params["closed"] not in ("right", "left"):
            raise PlotnineError("`closed` should either 'right' or 'left'")

        if (
            params["breaks"] is None
            and params["binwidth"] is None
            and params["bins"] is None
        ):
            params = params.copy()
            params["bins"] = freedman_diaconis_bins(data["x"])
            msg = (
                "'stat_bin()' using 'bins = {}'. "
                "Pick better value with 'binwidth'."
            )
            warn(msg.format(params["bins"]), PlotnineWarning)

        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        if params["breaks"] is not None:
            breaks = np.asarray(params["breaks"])
            if hasattr(scales.x, "transform"):
                breaks = scales.x.transform(breaks)
        elif params["binwidth"] is not None:
            breaks = breaks_from_binwidth(
                scales.x.dimension(),
                params["binwidth"],
                params["center"],
                params["boundary"],
            )
        else:
            breaks = breaks_from_bins(
                scales.x.dimension(),
                params["bins"],
                params["center"],
                params["boundary"],
            )

        new_data = assign_bins(
            data["x"],
            breaks,
            data.get("weight"),
            params["pad"],
            params["closed"],
        )
        return new_data



from contextlib import suppress
from copy import copy

from ..exceptions import PlotnineError
from ..mapping.aes import SCALED_AESTHETICS
from ..utils import jitter, resolution
from .position import position
from .position_dodge import position_dodge


# Adjust position by simultaneously dodging and jittering
class position_jitterdodge(position):
    """
    Dodge and jitter to minimise overlap

    Useful when aligning points generated through
    :class:`~plotnine.geoms.geom_point` with dodged a
    :class:`~plotnine.geoms.geom_boxplot`.

    Parameters
    ----------
    jitter_width : float
        Proportion to jitter in horizontal direction.
        Default is ``0.4`` of the resolution of the data.
    jitter_height : float
        Proportion to jitter in vertical direction.
        Default is ``0.0`` of the resolution of the data.
    dodge_width : float
        Amount to dodge in horizontal direction.
        Default is ``0.75``
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.
    """

    REQUIRED_AES = {"x", "y"}
    strategy = staticmethod(position_dodge.strategy)

    def __init__(
        self,
        jitter_width=None,
        jitter_height=0,
        dodge_width=0.75,
        random_state=None,
    ):
        self.params = {
            "jitter_width": jitter_width,
            "jitter_height": jitter_height,
            "dodge_width": dodge_width,
            "random_state": random_state,
        }

    def setup_params(self, data):
        params = copy(self.params)
        width = params["jitter_width"]
        if width is None:
            width = resolution(data["x"]) * 0.4

        # Adjust the x transformation based on the number
        # of dodge variables
        dvars = SCALED_AESTHETICS - self.REQUIRED_AES
        dodge_columns = data.columns.intersection(list(dvars))
        if len(dodge_columns) == 0:
            raise PlotnineError(
                "'position_jitterdodge' requires at least one "
                "aesthetic to dodge by."
            )

        s = set()
        for col in dodge_columns:
            with suppress(AttributeError):
                s.update(data[col].cat.categories)
        ndodge = len(s)

        params["jitter_width"] = width / (ndodge + 2)
        params["width"] = params["dodge_width"]
        return params

    @classmethod
    def compute_panel(cls, data, scales, params):
        trans_x = None  # pyright: ignore
        trans_y = None  # pyright: ignore

        if params["jitter_width"] > 0:

            def trans_x(x):
                return jitter(
                    x,
                    amount=params["jitter_width"],
                    random_state=params["random_state"],
                )

        if params["jitter_height"] > 0:

            def trans_y(y):
                return jitter(
                    y,
                    amount=params["jitter_height"],
                    random_state=params["random_state"],
                )

        # dodge, then jitter
        data = cls.collide(data, params=params)
        data = cls.transform_position(data, trans_x, trans_y)
        return data



import numpy as np
import pandas as pd

from plotnine import (
    aes,
    after_stat,
    geom_bar,
    geom_col,
    geom_histogram,
    geom_text,
    ggplot,
    scale_x_sqrt,
)
from plotnine.stats.binning import freedman_diaconis_bins

from .conftest import layer_data

n = 10  # Some even number greater than 2

# ladder: 0 1 times, 1 2 times, 2 3 times, ...
df = pd.DataFrame(
    {
        "x": np.repeat(range(n + 1), range(n + 1)),
        "z": np.repeat(range(n // 2), range(3, n * 2, 4)),
    }
)


def test_bar_count():
    p = ggplot(df, aes("x")) + geom_bar(aes(fill="factor(z)"))

    assert p == "bar-count"


def test_col():
    # The color indicates reveals the edges and the stacking
    # that is going on.
    p = ggplot(df) + geom_col(aes("x", "z", fill="factor(z)"), color="black")

    assert p == "col"


def test_histogram_count():
    p = ggplot(df, aes("x")) + geom_histogram(aes(fill="factor(z)"), bins=n)

    assert p == "histogram-count"


def test_scale_transformed_breaks():
    df = pd.DataFrame({"x": np.repeat(range(1, 5), range(1, 5))})
    p = ggplot(df, aes("x")) + geom_histogram(breaks=[1, 2.5, 4])
    out1 = layer_data(p)
    out2 = layer_data(p + scale_x_sqrt())
    np.testing.assert_allclose(out1.xmin, [1, 2.5])
    np.testing.assert_allclose(out2.xmin, np.sqrt([1, 2.5]))


def test_stat_count_int():
    df = pd.DataFrame({"x": ["a", "b"], "weight": [1, 2]})

    p = (
        ggplot(df)
        + aes(x="x", weight="weight", fill="x")
        + geom_bar()
        + geom_text(aes(label=after_stat("count")), stat="count")
    )

    assert p == "stat-count-int"


def test_stat_count_float():
    df = pd.DataFrame({"x": ["a", "b"], "weight": [1.5, 2.5]})

    p = (
        ggplot(df)
        + aes(x="x", weight="weight", fill="x")
        + geom_bar()
        + geom_text(aes(label=after_stat("count")), stat="count")
    )

    assert p == "stat-count-float"


def test_freedman_diaconis_bins():
    a1 = np.arange(1, 98, dtype=float)
    a2 = np.arange(100, dtype=float)
    a2[[0, 99]] = np.nan
    iqr1 = freedman_diaconis_bins(a1)
    iqr2 = freedman_diaconis_bins(a2)
    assert iqr1 == iqr2



"""
Statistics
"""
from .stat_bin import stat_bin
from .stat_bin_2d import stat_bin2d, stat_bin_2d
from .stat_bindot import stat_bindot
from .stat_boxplot import stat_boxplot
from .stat_count import stat_count
from .stat_density import stat_density
from .stat_density_2d import stat_density_2d
from .stat_ecdf import stat_ecdf
from .stat_ellipse import stat_ellipse
from .stat_function import stat_function
from .stat_hull import stat_hull
from .stat_identity import stat_identity
from .stat_pointdensity import stat_pointdensity
from .stat_qq import stat_qq
from .stat_qq_line import stat_qq_line
from .stat_quantile import stat_quantile
from .stat_sina import stat_sina
from .stat_smooth import stat_smooth
from .stat_sum import stat_sum
from .stat_summary import stat_summary
from .stat_summary_bin import stat_summary_bin
from .stat_unique import stat_unique
from .stat_ydensity import stat_ydensity

__all__ = [
    "stat_count",
    "stat_bin",
    "stat_bin_2d",
    "stat_bin2d",
    "stat_bindot",
    "stat_boxplot",
    "stat_density",
    "stat_ecdf",
    "stat_ellipse",
    "stat_density_2d",
    "stat_function",
    "stat_hull",
    "stat_identity",
    "stat_pointdensity",
    "stat_qq",
    "stat_qq_line",
    "stat_quantile",
    "stat_sina",
    "stat_smooth",
    "stat_sum",
    "stat_summary",
    "stat_summary_bin",
    "stat_unique",
    "stat_ydensity",
]



"""
Plotnine Data
"""
from pathlib import Path

import pandas as pd
from pandas.api.types import CategoricalDtype

__all__ = [
    "diamonds",
    "economics",
    "economics_long",
    "faithful",
    "faithfuld",
    "huron",
    "luv_colours",
    "meat",
    "midwest",
    "mpg",
    "msleep",
    "mtcars",
    "pageviews",
    "presidential",
    "seals",
    "txhousing",
]

data_dir = Path(__file__).parent

mtcars = pd.read_csv(data_dir / "mtcars.csv")
meat = pd.read_csv(data_dir / "meat.csv", parse_dates=[0])
pageviews = pd.read_csv(data_dir / "pageviews.csv", parse_dates=[0])
huron = pd.read_csv(data_dir / "huron.csv")
seals = pd.read_csv(data_dir / "seals.csv")
economics = pd.read_csv(data_dir / "economics.csv", parse_dates=[0])
economics_long = pd.read_csv(data_dir / "economics_long.csv", parse_dates=[0])
presidential = pd.read_csv(data_dir / "presidential.csv", parse_dates=[1, 2])
txhousing = pd.read_csv(data_dir / "txhousing.csv")
luv_colours = pd.read_csv(data_dir / "luv_colours.csv")
faithfuld = pd.read_csv(data_dir / "faithfuld.csv")
faithful = pd.read_csv(data_dir / "faithful.csv")


# add factors
def _ordered_categories(df, categories):
    """
    Make the columns in df categorical

    Parameters
    ----------
    df : pd.Dataframe
        Dataframe
    categories: dict
        Of the form {str: list},
        where the key the column name and the value is
        the ordered category list
    """
    for col, cats in categories.items():
        df[col] = df[col].astype(CategoricalDtype(cats, ordered=True))
    return df


def _unordered_categories(df, columns):
    """Make the columns in df categorical"""
    for col in columns:
        df[col] = df[col].astype(CategoricalDtype(ordered=False))
    return df


diamonds = pd.read_csv(data_dir / "diamonds.csv")
categories = {
    "cut": ["Fair", "Good", "Very Good", "Premium", "Ideal"],
    "clarity": ["I1", "SI2", "SI1", "VS2", "VS1", "VVS2", "VVS1", "IF"],
    "color": ["D", "E", "F", "G", "H", "I", "J"],
}
diamonds = _ordered_categories(diamonds, categories)

midwest = pd.read_csv(data_dir / "midwest.csv")
midwest = _unordered_categories(midwest, ["category"])

mpg = pd.read_csv(data_dir / "mpg.csv")
columns = ["manufacturer", "model", "trans", "fl", "drv", "class"]
mpg = _unordered_categories(mpg, columns)

msleep = pd.read_csv(data_dir / "msleep.csv")
msleep = _unordered_categories(msleep, ["vore", "conservation"])


# Documentation

mtcars.__doc__ = """
Motor Trend Car Road Tests

.. rubric:: Description

The data was extracted from the 1974 *Motor Trend* US magazine,
and comprises fuel consumption and 10 aspects of automobile
design and performance for 32 automobiles (1973–74 models).


.. rubric:: Format

A data frame with 32 observations on 11 variables.

======  =========================================
Column  Description
======  =========================================
mpg     Miles/(US) gallon
cyl     Number of cylinders
disp    Displacement (cu.in.)
hp      Gross horsepower
drat    Rear axle ratio
wt      Weight (1000 lbs)
qsec    1/4 mile time
vs      V/S
am      Transmission (0 = automatic, 1 = manual)
gear    Number of forward gears
carb    Number of carburetors
======  =========================================

.. rubric:: Source

Henderson and Velleman (1981), Building multiple regression
models interactively. *Biometrics*, **37**, 391–411.
"""

meat.__doc__ = """

"""

pageviews.__doc__ = """
"""

huron.__doc__ = """
Level of Lake Huron 1875–1972

.. rubric:: Description

Annual measurements of the level, in feet, of Lake Huron 1875–1972.

.. rubric:: Format

=========   ==============
Column      Description
=========   ==============
year        Year
level       Water level
decade      Decade
=========   ==============

.. rubric:: Source

Brockwell, P. J. and Davis, R. A. (1991). Time Series and Forecasting Methods.
Second edition. Springer, New York. Series A, page 555.

Brockwell, P. J. and Davis, R. A. (1996). Introduction to Time Series and
Forecasting. Springer, New York. Sections 5.1 and 7.6.
"""

seals.__doc__ = """
Vector field of seal movements.

.. rubric:: Description

This vector field was produced from the data described in Brillinger,
D.R., Preisler, H.K., Ager, A.A. and Kie, J.G.
"An exploratory data analysis (EDA) of the paths of moving animals". J.
Statistical Planning and Inference 122 (2004), 43-63, using the methods
of Brillinger, D.R., "Learning a potential function from a trajectory",
Signal Processing Letters. December (2007).

.. rubric:: Format

A data frame with 1155 rows and 4 variables

===========  ===================
Column       Description
===========  ===================
lat          Latitude
long         Longitude
delta_long   Change in Longitude
delta_lat    Change in Latitude
===========  ===================

.. rubric:: References

http://www.stat.berkeley.edu/~brill/Papers/jspifinal.pdf
"""

economics.__doc__ = """
US economic time series.

.. rubric:: Description

This dataset was produced from US economic time series data available
from http://research.stlouisfed.org/fred2.
`economics` is in "wide" format, `economics_long` is in "long" format.

.. rubric:: Format

A data frame with 478 rows and 6 variables

=========  ==========================================================
Column     Description
=========  ==========================================================
date       Month of data collection
psavert    personal savings rate [1_]
pce        personal consumption expenditures, in billions of dollars [2_]
unemploy   number of unemployed in thousands, [3_]
uempmed    median duration of unemployment, in week [4_]
pop        total population, in thousands [5_]
=========  ==========================================================

.. _1: http://research.stlouisfed.org/fred2/series/PSAVERT/
.. _2: http://research.stlouisfed.org/fred2/series/PCE
.. _3: http://research.stlouisfed.org/fred2/series/UNEMPLOY
.. _4: http://research.stlouisfed.org/fred2/series/UEMPMED
.. _5: http://research.stlouisfed.org/fred2/series/POP
"""

economics_long.__doc__ = economics.__doc__

presidential.__doc__ = """
Terms of 11 presidents from Eisenhower to Obama.

.. rubric:: Description

The names of each president, the start and end date
of their term, and their party of 11 US presidents
from Eisenhower to Obama.

==========     ===========================
Column         Description
==========     ===========================
name           Name of president
start          Start of presidential term
end            End of presidential term
party          Political Party
==========     ===========================

.. rubric:: Format

A data frame with 11 rows and 4 variables
"""

txhousing.__doc__ = """
Housing sales in TX.

.. rubric:: Description

Information about the housing market in Texas provided
by the TAMU real estate center, http://recenter.tamu.edu/.


.. rubric:: Format

A data frame with 8602 observations and 9 variables:

=========   ===============================================
Column      Description
=========   ===============================================
city        Name of MLS area
year        Year
month       Month
sales       Number of sales
volume      Total value of sales
median      Median sale price
listings    Total active listings
inventory   "Months inventory": amount of time it would \n
            take to sell all current listings at current \n
            pace of sales.
date        Date
=========   ===============================================
"""

luv_colours.__doc__ = """
colors in Luv space.

.. rubric:: Description

Named colors translated into Luv colour space.

luv_colours
.. rubric:: Format

A data frame with 657 observations and 4 variables:

======  ============================
Column  Description
======  ============================
L       L position in Luv colour space
u       u position in Luv colour space
v       v position in Luv colour space
col     Colour name
======  ============================
"""

faithful.__doc__ = """
Old Faithful Geyser Data

.. rubric:: Description

Waiting time between eruptions and the duration of the
eruption for the Old Faithful geyser in Yellowstone
National Park, Wyoming, USA.

.. rubric:: Format

A data frame with 272 observations on 2 variables.

==========   ========================================
Column       Description
==========   ========================================
eruptions    Eruption time in mins
waiting	     Waiting time to next eruption (in mins)
==========   ========================================

.. rubric:: Details

A closer look at `faithful.eruptions` reveals that these are
heavily rounded times originally in seconds, where multiples
of 5 are more frequent than expected under non-human measurement.
For a better version of the eruption times, see the example below.

There are many versions of this dataset around:
Azzalini and Bowman (1990) use a more complete version.

.. rubric:: Source

W. Härdle.

.. rubric:: References

Härdle, W. (1991) *Smoothing Techniques with Implementation in S*.
New York: Springer.

Azzalini, A. and Bowman, A. W. (1990). A look at some data
on the Old Faithful geyser. **Applied Statistics** *39*, 357–365.
"""

faithfuld.__doc__ = """
Old Faithful Geyser Data

.. rubric:: Description

Waiting time between eruptions and the duration of the
eruption for the Old Faithful geyser in Yellowstone
National Park, Wyoming, USA.

.. rubric:: Format

A data frame with *grid data* for 272 observations on 2
variables and the density at those locations.

==========   ========================================
Column       Description
==========   ========================================
eruptions    Eruption time in mins
waiting	     Waiting time to next eruption (in mins)
density	     Density Estimate
==========   ========================================

.. rubric:: Details

A closer look at `faithful.eruptions` reveals that these are
heavily rounded times originally in seconds, where multiples
of 5 are more frequent than expected under non-human measurement.
For a better version of the eruption times, see the example below.

There are many versions of this dataset around:
Azzalini and Bowman (1990) use a more complete version.

.. rubric:: Source

W. Härdle.

.. rubric:: References

Härdle, W. (1991) *Smoothing Techniques with Implementation in S*.
New York: Springer.

Azzalini, A. and Bowman, A. W. (1990). A look at some data
on the Old Faithful geyser. **Applied Statistics** *39*, 357–365.
"""

diamonds.__doc__ = """
Prices of 50,000 round cut diamonds

.. rubric:: Description

A dataset containing the prices and other attributes
of almost 54,000 diamonds. The variables are as follows:

.. rubric:: Format

A data frame with 53940 rows and 10 variables:

========  ==================================
Column    Description
========  ==================================
price     price in US dollars ($326–$18,823)
carat     weight of the diamond (0.2–5.01)
cut       quality of the cut (Fair, Good, Very Good, Premium, Ideal)
color     diamond colour, from J (worst) to D (best)
clarity   a measurement of how clear the diamond is \n
          (I1 (worst), SI1, SI2, VS1, VS2, VVS1, VVS2, IF (best))
x         length in mm (0–10.74)
y         width in mm (0–58.9)
z         depth in mm (0–31.8)
depth     total depth percentage = z / mean(x, y) = 2 * z / (x + y) (43–79)
table     width of top of diamond relative to widest point (43–95)
========  ==================================
"""

midwest.__doc__ = """
Midwest demographics.

.. rubric:: Description

Demographic information of midwest counties

.. rubric:: Format

A data frame with 437 rows and 28 variables

=====================  ============================
Column                 Description
=====================  ============================
PID
county
state
area
poptotal               Total population
popdensity             Population density
popwhite               Number of whites
popblack               Number of blacks
popamerindian          Number of American Indians
popasian               Number of Asians
popother               Number of other races
percwhite              Percent white
percblack              Percent black
percamerindan          Percent American Indian
percasian              Percent Asian
percother              Percent other races
popadults              Number of adults
perchsd
percollege             Percent college educated
percprof               Percent profession
poppovertyknown
percpovertyknown
percbelowpoverty
percchildbelowpovert
percadultpoverty
percelderlypoverty
inmetro                In a metro area
category
=====================  ============================
"""

mpg.__doc__ = """
Fuel economy data from 1999 and 2008 for 38 popular models of car

.. rubric:: Description

This dataset contains a subset of the fuel economy data that
the EPA makes available on http://fueleconomy.gov.
It contains only models which had a new release every year
between 1999 and 2008 - this was used as a proxy for
the popularity of the car.

.. rubric:: Format

A data frame with 234 rows and 11 variables

============   ====================================================
Column         Description
============   ====================================================
manufacturer
model
displ          engine displacement, in litres
year
cyl            number of cylinders
trans          type of transmission
drv            f = front-wheel drive, r = rear wheel drive, 4 = 4wd
cty            city miles per gallon
hwy            highway miles per gallon
fl
class
============   ====================================================
"""

msleep.__doc__ = """
An updated and expanded version of the mammals sleep dataset.

.. rubric:: Description

This is an updated and expanded version of the mammals
sleep dataset. Updated sleep times and weights were taken
from V. M. Savage and G. B. West. A quantitative, theoretical
framework for understanding mammalian sleep. Proceedings of
the National Academy of Sciences, 104 (3):1051-1056, 2007.

.. rubric:: Format

A data frame with 83 rows and 11 variables

=============  =====================================
Column         Description
=============  =====================================
name           common name
genus
vore           carnivore, omnivore or herbivore?
order
conservation   the conservation status of the animal
sleep_total    total amount of sleep, in hours
sleep_rem      rem sleep, in hours
sleep_cycle    length of sleep cycle, in hours
awake          amount of time spent awake, in hours
brainwt        brain weight in kilograms
bodywt         body weight in kilograms
=============  =====================================

.. rubric:: Details

Additional variables order, conservation status and
vore were added from wikipedia.
"""



from __future__ import annotations

import typing

from ..doctools import document
from ..utils import resolution
from .geom_rect import geom_rect

if typing.TYPE_CHECKING:
    import pandas as pd


@document
class geom_bar(geom_rect):
    """
    Bar plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float, optional (default None)
        Bar width. If :py:`None`, the width is set to
        `90%` of the resolution of the data.

    See Also
    --------
    plotnine.geoms.geom_histogram
    """

    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"xmin", "xmax", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "count",
        "position": "stack",
        "na_rm": False,
        "width": None,
    }

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if "width" not in data:
            if self.params["width"]:
                data["width"] = self.params["width"]
            else:
                data["width"] = resolution(data["x"], False) * 0.9

        bool_idx = data["y"] < 0

        data["ymin"] = 0.0
        data.loc[bool_idx, "ymin"] = data.loc[bool_idx, "y"]

        data["ymax"] = data["y"]
        data.loc[bool_idx, "ymax"] = 0.0

        data["xmin"] = data["x"] - data["width"] / 2
        data["xmax"] = data["x"] + data["width"] / 2
        del data["width"]
        return data



import numpy as np
import pandas as pd

from plotnine import aes, ggplot, stat_summary_bin

df = pd.DataFrame(
    {
        "xd": list("aaaaabbbbcccccc"),
        "xc": range(0, 15),
        "y": [1, 2, 3, 4, 5, 1.5, 1.5, 6, 6, 5, 5, 5, 5, 5, 5],
    }
)


def test_discrete_x():
    p = ggplot(df, aes("xd", "y")) + stat_summary_bin(
        fun_y=np.mean, fun_ymin=np.min, fun_ymax=np.max, geom="bar"
    )

    assert p == "discrete_x"


def test_continuous_x():
    p = ggplot(df, aes("xc", "y")) + stat_summary_bin(
        fun_y=np.mean, fun_ymin=np.min, fun_ymax=np.max, bins=5, geom="bar"
    )

    assert p == "continuous_x"



from contextlib import suppress
from copy import copy

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError
from ..utils import groupby_apply, match
from .position import position


class position_dodge(position):
    """
    Dodge overlaps and place objects side-by-side

    Parameters
    ----------
    width: float
        Dodging width, when different to the width of the
        individual elements. This is useful when you want
        to align narrow geoms with wider geoms
    preserve: str in ``['total', 'single']``
        Should dodging preserve the total width of all elements
        at a position, or the width of a single element?
    """

    REQUIRED_AES = {"x"}

    def __init__(self, width=None, preserve="total"):
        self.params = {
            "width": width,
            "preserve": preserve,
        }

    def setup_data(self, data, params):
        has_xmin_xmax = "xmin" in data and "xmax" in data
        if "x" not in data and has_xmin_xmax:
            data["x"] = (data["xmin"] + data["xmax"]) / 2
        return super().setup_data(data, params)

    def setup_params(self, data):
        if (
            ("xmin" not in data)
            and ("xmax" not in data)
            and (self.params["width"] is None)
        ):
            msg = "Width not defined. " "Set with `position_dodge(width = ?)`"
            raise PlotnineError(msg)

        params = copy(self.params)

        if params["preserve"] == "total":
            params["n"] = None
        else:
            # Count at the xmin values per panel and find the highest
            # overall count
            def max_xmin_values(gdf):
                try:
                    n = gdf["xmin"].value_counts().max()
                except KeyError:
                    n = gdf["x"].value_counts().max()
                return pd.DataFrame({"n": [n]})

            res = groupby_apply(data, "PANEL", max_xmin_values)
            params["n"] = res["n"].max()
        return params

    @classmethod
    def compute_panel(cls, data, scales, params):
        return cls.collide(data, params=params)

    @staticmethod
    def strategy(data, params):
        """
        Dodge overlapping interval

        Assumes that each set has the same horizontal position.
        """
        width = params["width"]
        with suppress(TypeError):
            iter(width)
            width = np.asarray(width)
            width = width[data.index]

        udata_group = data["group"].drop_duplicates()

        n = params.get("n", None)
        if n is None:
            n = len(udata_group)
        if n == 1:
            return data

        if not all(col in data.columns for col in ["xmin", "xmax"]):
            data["xmin"] = data["x"]
            data["xmax"] = data["x"]

        d_width = np.max(data["xmax"] - data["xmin"])

        # Have a new group index from 1 to number of groups.
        # This might be needed if the group numbers in this set don't
        # include all of 1:n
        udata_group = udata_group.sort_values()
        groupidx = match(data["group"], udata_group)
        groupidx = np.asarray(groupidx) + 1

        # Find the center for each group, then use that to
        # calculate xmin and xmax
        data["x"] = data["x"] + width * ((groupidx - 0.5) / n - 0.5)
        data["xmin"] = data["x"] - (d_width / n) / 2
        data["xmax"] = data["x"] + (d_width / n) / 2

        return data



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..mapping import aes
from ..utils import SIZE_FACTOR, order_as_data_mapping, to_rgba
from .geom import geom
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Aes,
        Axes,
        Coord,
        DataLike,
        DrawingArea,
        Layer,
    )


@document
class geom_vline(geom):
    """
    Vertical line

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
        "alpha": 1,
    }
    REQUIRED_AES = {"xintercept"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "inherit_aes": False,
    }

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        data, mapping = order_as_data_mapping(data, mapping)
        xintercept = kwargs.pop("xintercept", None)
        if xintercept is not None:
            if mapping:
                warn(
                    "The 'xintercept' parameter has overridden "
                    "the aes() mapping.",
                    PlotnineWarning,
                )
            data = pd.DataFrame({"xintercept": np.repeat(xintercept, 1)})
            mapping = aes(xintercept="xintercept")
            kwargs["show_legend"] = False

        geom.__init__(self, mapping, data, **kwargs)

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        ranges = coord.backtransform_range(panel_params)
        data["x"] = data["xintercept"]
        data["xend"] = data["xintercept"]
        data["y"] = ranges.y[0]
        data["yend"] = ranges.y[1]
        data = data.drop_duplicates()

        for _, gdata in data.groupby("group"):
            gdata.reset_index(inplace=True)
            geom_segment.draw_group(gdata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a vertical line in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D

        x = [0.5 * da.width] * 2
        y = [0, da.height]
        data["size"] *= SIZE_FACTOR
        color = to_rgba(data["color"], data["alpha"])
        key = Line2D(
            x,
            y,
            linestyle=data["linetype"],
            linewidth=data["size"],
            color=color,
            solid_capstyle="butt",
            antialiased=False,
        )
        da.add_artist(key)
        return da



from ..doctools import document
from .geom_path import geom_path


@document
class geom_qq_line(geom_path):
    """
    Quantile-Quantile Line plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {
        "stat": "qq_line",
        "position": "identity",
        "na_rm": False,
    }



import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..scales.scale import scale_discrete
from ..utils import groupby_apply
from .binning import fuzzybreaks
from .stat import stat
from .stat_summary import make_summary_fun


@document
class stat_summary_bin(stat):
    """
    Summarise y values at x intervals

    {usage}

    Parameters
    ----------
    {common_parameters}
    binwidth : float or tuple, optional (default: None)
        The width of the bins. The default is to use bins bins that
        cover the range of the data. You should always override this
        value, exploring multiple widths to find the best to illustrate
        the stories in your data.
    bins : int or tuple, optional (default: 30)
        Number of bins. Overridden by binwidth.
    breaks : array-like(s), optional (default: None)
        Bin boundaries. This supercedes the ``binwidth``, ``bins``
        and ``boundary`` arguments.
    boundary : float or tuple, optional (default: None)
        A boundary between two bins. As with center, things are
        shifted when boundary is outside the range of the data.
        For example, to center on integers, use :py:`width=1` and
        :py:`boundary=0.5`, even if 1 is outside the range of the
        data. At most one of center and boundary may be specified.
    fun_data : str or function, optional
        One of
        :py:`['mean_cl_boot', 'mean_cl_normal', 'mean_sdl', 'median_hilow',
        'mean_se']` or any function that takes a array and returns a
        dataframe with three rows indexed as ``y``, ``ymin`` and ``ymax``.
        Defaults to :py:`'mean_se'`.
    fun_y : function, optional (default: None)
        Any function that takes an array-like and returns a single value
    fun_ymax : function, optional (default: None)
        Any function that takes an array-like and returns a single value
    fun_args : dict, optional (default: None)
        Arguments to any of the functions. Provided the names of the
        arguments of the different functions are in not conflict, the
        arguments will be assigned to the right functions. If there is
        a conflict, create a wrapper function that resolves the
        ambiguity in the argument names.
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.

    Notes
    -----
    The *binwidth*, *bins*, *breaks* and *bounary* arguments can be a
    tuples with two values (``(xaxis-value, yaxis-value)``) of the
    required type.

    See Also
    --------
    plotnine.geoms.geom_pointrange
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'bin'    # bin identifier
        'width'  # bin width
        'ymin'   # ymin computed by the summary function
        'ymax'   # ymax computed by the summary function

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('ymin')`.
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "pointrange",
        "position": "identity",
        "na_rm": False,
        "bins": 30,
        "breaks": None,
        "binwidth": None,
        "boundary": None,
        "fun_data": None,
        "fun_y": None,
        "fun_ymin": None,
        "fun_ymax": None,
        "fun_args": None,
        "random_state": None,
    }
    CREATES = {"bin", "width", "ymin", "ymax"}

    def setup_params(self, data):
        keys = ("fun_data", "fun_y", "fun_ymin", "fun_ymax")
        if not any(self.params[k] for k in keys):
            PlotnineWarning(
                "No summary function, supplied, defaulting to mean_se()"
            )
            self.params["fun_data"] = "mean_se"

        if self.params["fun_args"] is None:
            self.params["fun_args"] = {}

        if "random_state" not in self.params["fun_args"]:
            if self.params["random_state"]:
                random_state = self.params["random_state"]
                if random_state is None:
                    random_state = np.random
                elif isinstance(random_state, int):
                    random_state = np.random.RandomState(random_state)

                self.params["fun_args"]["random_state"] = random_state

        return self.params

    @classmethod
    def compute_group(cls, data, scales, **params):
        bins = params["bins"]
        breaks = params["breaks"]
        binwidth = params["binwidth"]
        boundary = params["boundary"]

        func = make_summary_fun(
            params["fun_data"],
            params["fun_y"],
            params["fun_ymin"],
            params["fun_ymax"],
            params["fun_args"],
        )

        breaks = fuzzybreaks(scales.x, breaks, boundary, binwidth, bins)
        data["bin"] = pd.cut(
            data["x"],  # pyright: ignore
            bins=breaks,  # pyright: ignore
            labels=False,  # pyright: ignore
            include_lowest=True,
        )

        def func_wrapper(data: pd.DataFrame) -> pd.DataFrame:
            """
            Add `bin` column to each summary result.
            """
            result = func(data)
            result["bin"] = data["bin"].iloc[0]
            return result

        # This is a plyr::ddply
        out = groupby_apply(data, "bin", func_wrapper)
        centers = (breaks[:-1] + breaks[1:]) * 0.5
        bin_centers = centers[out["bin"].to_numpy()]
        out["x"] = bin_centers
        out["bin"] += 1
        if isinstance(scales.x, scale_discrete):
            out["width"] = 0.9
        else:
            out["width"] = np.diff(breaks)[bins - 1]

        return out



import pandas as pd
import pytest

from plotnine import aes, geom_point, geom_vline, ggplot
from plotnine.exceptions import PlotnineError, PlotnineWarning

df = pd.DataFrame(
    {"xintercept": [1, 2], "x": [0.5, 3], "y": [-1, 1], "z": range(2)}
)


def test_aesthetics():
    p = (
        ggplot(df)
        + geom_point(aes("x", "y"))
        + geom_vline(aes(xintercept="xintercept"), size=2)
        + geom_vline(aes(xintercept="xintercept+.1", alpha="z"), size=2)
        + geom_vline(
            aes(xintercept="xintercept+.2", linetype="factor(z)"), size=2
        )
        + geom_vline(
            aes(xintercept="xintercept+.3", color="factor(z)"), size=2
        )
        + geom_vline(aes(xintercept="xintercept+.4", size="z"))
    )

    assert p == "aesthetics"


def test_aes_inheritance():
    with pytest.raises(PlotnineError):
        p = (
            ggplot(df, aes("x", "y", xintercept="xintercept"))
            + geom_point()
            + geom_vline(size=2)
        )
        p.draw_test()


def test_aes_overwrite():
    with pytest.warns(PlotnineWarning):
        geom_vline(aes(color="x"), xintercept=2)



import warnings
from datetime import datetime

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from plotnine import (
    aes,
    annotate,
    element_text,
    expand_limits,
    facet_wrap,
    geom_bar,
    geom_col,
    geom_point,
    ggplot,
    lims,
    theme,
)
from plotnine.exceptions import PlotnineError, PlotnineWarning
from plotnine.scales import (
    scale_color,
    scale_color_manual,
    scale_identity,
    scale_manual,
    scale_xy,
)
from plotnine.scales.scale_alpha import (
    scale_alpha_continuous,
    scale_alpha_discrete,
)
from plotnine.scales.scale_linetype import (
    scale_linetype_continuous,
    scale_linetype_discrete,
)
from plotnine.scales.scale_manual import _scale_manual
from plotnine.scales.scale_shape import (
    scale_shape_continuous,
    scale_shape_discrete,
)
from plotnine.scales.scale_size import (
    scale_size_area,
    scale_size_continuous,
    scale_size_discrete,
    scale_size_radius,
)
from plotnine.scales.scale_xy import (
    scale_x_continuous,
    scale_x_discrete,
    scale_y_continuous,
)
from plotnine.scales.scales import make_scale


# test palettes
def test_discrete_color_palettes():
    def _assert_all_colors(colors, n, m=0):
        """
        Make sure the first n elements of colors
        are rgb hex strings. And that the last
        m elements are None
        """
        assert len(colors) == n + m
        assert all(c.startswith("#") for c in colors[:n])
        if m > 0:
            assert all(c is None for c in colors[-m:])

    sc = scale_color

    s = sc.scale_color_hue()
    colors = s.palette(5)
    _assert_all_colors(colors, 5)

    s = sc.scale_color_grey()
    colors = s.palette(5)
    _assert_all_colors(colors, 5)

    # sequential palettes have a maximum of 9 colors
    s = sc.scale_color_brewer(type="seq")
    colors = s.palette(5)
    _assert_all_colors(colors, 5)

    s = sc.scale_color_brewer(type="seq")
    colors = s.palette(9)
    _assert_all_colors(colors, 9)

    s = sc.scale_color_brewer(type="seq")
    with pytest.warns(UserWarning):  # upstream warning
        colors = s.palette(15)
    _assert_all_colors(colors, 9, 6)

    # diverging palettes have a maximum of 11 colors
    s = sc.scale_color_brewer(type="div")
    colors = s.palette(5)
    _assert_all_colors(colors, 5)

    s = sc.scale_color_brewer(type="div")
    colors = s.palette(11)
    _assert_all_colors(colors, 11)

    s = sc.scale_color_brewer(type="div")
    with pytest.warns(UserWarning):  # upstream warning
        colors = s.palette(21)
    _assert_all_colors(colors, 11, 10)

    # qualitative have varying maximum colors
    s = sc.scale_color_brewer(type="qual")
    colors = s.palette(5)
    _assert_all_colors(colors, 5)

    s = sc.scale_color_brewer(type="qual", palette="Accent")
    with pytest.warns(UserWarning):  # upstream warning
        colors = s.palette(12)
    _assert_all_colors(colors, 8, 4)

    s = sc.scale_color_brewer(type="qual", palette="Set3")
    with pytest.warns(UserWarning):  # upstream warning
        colors = s.palette(15)
    _assert_all_colors(colors, 12, 3)


def test_continuous_color_palettes():
    alpha = 0.6
    alphas = [0.1, 0.9, 0.32, 1.0, 0.65]
    colors1 = ["#000000", "#11BB20"]
    colors2 = ["#000000", "#003399", "#42BF63", "#191141"]
    sc = scale_color

    def _assert(cscale):
        """
        Make color scale palette returns a single color when
        passed a scalar and multiple colors when
        passed a list
        """
        color = cscale.palette(alpha)
        assert color[0] == "#"

        colors = cscale.palette(alphas)
        assert all(c[0] == "#" for c in colors)

    s = sc.scale_color_gradient()
    _assert(s)

    s = sc.scale_color_gradient2()
    _assert(s)

    s = sc.scale_color_gradientn(colors1)
    _assert(s)

    s = sc.scale_color_gradientn(colors2)
    _assert(s)

    s = sc.scale_color_distiller(type="seq")
    _assert(s)

    s = sc.scale_color_distiller(type="div")
    _assert(s)

    with pytest.warns(PlotnineWarning):
        s = sc.scale_color_distiller(type="qual")
    _assert(s)


def test_color_aliases():
    # American and British names should refer to the same scales
    names = (
        (s, s.replace("color", "colour"))
        for s in dir(scale_color)
        if s.startswith("scale_color")
    )

    for a, b in names:
        assert getattr(scale_color, a) is getattr(scale_color, b)


def test_fill_scale_aesthetics():
    for name in scale_color.__dict__:
        if name.startswith("scale_fill"):
            scale = getattr(scale_color, name)
            assert scale._aesthetics == ["fill"]


def test_linetype_palettes():
    N = 4  # distinct linetypes
    s = scale_linetype_discrete()
    items = s.palette(N)
    assert len(items) == N
    assert all(isinstance(x, str) for x in items)

    with pytest.warns(UserWarning):
        # More values than palette has
        items = s.palette(N + 5)
        assert all(isinstance(x, str) for x in items[:N])

    with pytest.raises(PlotnineError):
        s = scale_linetype_continuous()


def test_shape_palettes():
    N = 10  # distinct shapes
    s = scale_shape_discrete()
    items = s.palette(N)
    assert len(items) == N
    assert all(isinstance(x, str) for x in items)

    with pytest.warns(UserWarning):
        # More values than palette has
        items = s.palette(N + 5)
        assert all(isinstance(x, str) for x in items[:N])

    with pytest.raises(PlotnineError):
        scale_shape_continuous()

    # Unfilled
    N = 14
    values = list(range(14))
    s = scale_shape_discrete(unfilled=True)
    items = s.map(values, limits=values)
    assert len(items) == N
    assert not all(isinstance(x, str) for x in items)


def test_size_palette():
    with pytest.warns(PlotnineWarning):
        # Warns against a discrete size scale
        s = scale_size_discrete()
        items = s.palette(9)
        assert len(items) == 9

    s = scale_size_continuous(range=(1, 6))
    frac = 0.5
    value = s.palette(frac**2)
    assert value == (1 + 6) * frac

    # Just test that they work
    s = scale_size_area(max_size=6)
    s.palette(frac**2)

    s = scale_size_radius(range=(1, 6))
    s.palette(frac**2)


def test_scale_identity():
    def is_identity_scale(name):
        return name.startswith("scale_") and name.endswith("_identity")

    for name in scale_identity.__dict__:
        if is_identity_scale(name):
            s = getattr(scale_identity, name)()
            assert s.map([1, 2, 3]) == [1, 2, 3]


def test_scale_manual():
    def is_manual_scale(name):
        return name.startswith("scale_") and name.endswith("_manual")

    manual_scales = [
        getattr(scale_manual, name)
        for name in scale_manual.__dict__
        if is_manual_scale(name)
    ]

    values = [1, 2, 3, 4, 5]
    for _scale in manual_scales:
        s = _scale(values)
        assert s.palette(2) == values
        assert s.palette(len(values)) == values
        with pytest.warns(PlotnineWarning):
            s.palette(len(values) + 1)

    values = {"A": "red", "B": "violet", "C": "blue"}
    sc = scale_manual.scale_color_manual(values)
    assert sc.palette(3) == values

    # Breaks are matched with values
    sc1 = scale_manual.scale_color_manual(
        breaks=[True, False], values=["blue", "red"]
    )
    sc2 = scale_manual.scale_color_manual(
        breaks=[True, False], values=["red", "blue"]
    )
    assert sc1.map([True, False, True, False]) == ["blue", "red"] * 2
    assert sc2.map([True, False, True, False]) == ["red", "blue"] * 2


def test_alpha_palette():
    with pytest.warns(PlotnineWarning):
        s = scale_alpha_discrete()
        items = s.palette(9)
        assert len(items) == 9

    s = scale_alpha_continuous(range=(0.1, 1))
    value = s.palette(0.5)
    assert value == (0.1 + 1) / 2.0


def test_xy_limits():
    lst = [1, 2, 3]
    arr = np.array(lst)
    series = pd.Series(lst)
    s1 = scale_x_discrete(limits=lst)
    s2 = scale_x_discrete(limits=arr)
    s3 = scale_x_discrete(limits=series)
    assert all(s2.limits == s1.limits)
    assert all(s3.limits == s1.limits)


def test_setting_limits():
    lst = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    s = scale_x_continuous()
    s.train(lst)
    assert s.limits == (1, 10)

    s = scale_x_continuous(limits=(3, 7))
    s.train(lst)
    assert s.limits == (3, 7)

    s = scale_x_continuous(limits=(3, None))
    s.train(lst)
    assert s.limits == (3, 10)

    s = scale_x_continuous(limits=(None, 7))
    s.train(lst)
    assert s.limits == (1, 7)

    s = scale_color.scale_color_hue(limits=tuple("abcdefg"))
    s.train(["a", "b", "a"])
    assert s.limits == tuple("abcdefg")


def test_discrete_xy_scale_limits():
    lst = list("abcd")
    x = pd.Series(pd.Categorical(lst, ordered=True))

    s = scale_x_discrete()
    s.train(x)
    assert s.limits == lst

    s = scale_x_discrete(limits=reversed)
    s.train(x)
    assert s.limits == lst[::-1]


def test_discrete_xy_scale_drop_limits():
    df = pd.DataFrame(
        {"x": list("aaaabbbbccccddd"), "c": list("112312231233123")}
    )

    p = (
        ggplot(df)
        + geom_bar(aes(x="x", fill="c"))
        + scale_x_discrete(limits=list("abc"))
    )
    with pytest.warns(PlotnineWarning):
        assert p == "discrete_xy_scale_drop_limits"


def test_setting_limits_transformed():
    lst = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    s = scale_y_continuous(trans="log10")
    s.train(lst)
    assert s.limits == (1, 10)

    s = scale_y_continuous(trans="log10", limits=[2, 7])
    s.train(lst)
    assert s.limits == (np.log10(2), np.log10(7))

    s = scale_y_continuous(trans="log10", limits=[2, None])
    s.train(lst)
    assert s.limits == (np.log10(2), np.log10(10))

    s = scale_y_continuous(trans="log10", limits=[None, 7])
    s.train(lst)
    assert s.limits == (np.log10(1), np.log10(7))


def test_scale_continuous_limits_as_function():
    def reverse(x):
        return list(reversed(x))

    x = [1, 2, 3, 4]
    sc1 = scale_x_continuous()
    sc2 = scale_x_continuous(limits=reverse)
    sc1.train(x)
    sc2.train(x)
    assert sc1.limits == sc2.limits[::-1]


def test_scale_discrete_limits_as_function():
    def reverse(x):
        return list(reversed(x))

    x = ["a", "b", "c", "d"]
    sc1 = scale_color.scale_color_discrete()
    sc2 = scale_color.scale_color_discrete(limits=reverse)
    sc1.train(x)
    sc2.train(x)
    assert sc1.limits == sc2.limits[::-1]


def test_minor_breaks():
    n = 10
    x = np.arange(n)

    # Default
    s = scale_x_continuous()
    s.train(x)
    breaks = s.get_breaks()
    minor_breaks = s.get_minor_breaks(breaks)
    expected_minor_breaks = (breaks[:-1] + breaks[1:]) / 2
    assert np.allclose(minor_breaks, expected_minor_breaks, rtol=1e-12)

    # List
    expected_minor_breaks = [2, 4, 6, 8]
    s = scale_x_continuous(minor_breaks=expected_minor_breaks)
    s.train(x)
    breaks = s.get_breaks()
    minor_breaks = s.get_minor_breaks(breaks)
    assert np.allclose(minor_breaks, expected_minor_breaks, rtol=1e-12)

    # Callable
    def func(limits):
        return np.linspace(limits[0], limits[1], n)

    s = scale_x_continuous(minor_breaks=func)
    s.train(x)
    breaks = s.get_breaks()
    minor_breaks = s.get_minor_breaks(breaks)
    _breaks = set(breaks)
    expected_minor_breaks = [x for x in np.arange(n) if x not in _breaks]
    assert np.allclose(minor_breaks, expected_minor_breaks, rtol=1e-12)
    assert not (_breaks & set(minor_breaks))

    # Number of minor breaks
    s = scale_x_continuous(limits=[0, 20], minor_breaks=3)
    minor_breaks = s.get_minor_breaks(major=[0, 10, 20])
    expected_minor_breaks = [2.5, 5, 7.5, 12.5, 15, 17.5]
    assert np.allclose(minor_breaks, expected_minor_breaks, rtol=1e-12)


def test_expand_limits():
    df = pd.DataFrame({"x": range(5, 11), "y": range(5, 11)})
    p = ggplot(df, aes("x", "y")) + geom_point() + expand_limits(y=(0, None))
    assert p == "expand_limits"


def test_bool_mapping():
    df = pd.DataFrame({"x": [1, 2, 3], "y": [True, False, False]})
    p = ggplot(df, aes("x", "y")) + geom_point()
    assert p == "bool_mapping"


def test_make_scale_and_datetimes():
    def correct_scale(scale, name):
        return scale.__class__.__name__ == name

    # cpython
    x = pd.Series([datetime(year, 1, 1) for year in [2010, 2026, 2015]])

    assert correct_scale(make_scale("x", x), "scale_x_datetime")
    assert correct_scale(make_scale("color", x), "scale_color_datetime")
    assert correct_scale(make_scale("fill", x), "scale_fill_datetime")
    assert correct_scale(make_scale("size", x), "scale_size_datetime")
    assert correct_scale(make_scale("alpha", x), "scale_alpha_datetime")

    # numpy
    x = pd.Series([np.datetime64(i * 10, "D") for i in range(1, 10)])
    assert correct_scale(make_scale("x", x), "scale_x_datetime")
    assert correct_scale(make_scale("color", x), "scale_color_datetime")
    assert correct_scale(make_scale("fill", x), "scale_fill_datetime")
    assert correct_scale(make_scale("size", x), "scale_size_datetime")
    assert correct_scale(make_scale("alpha", x), "scale_alpha_datetime")


def test_scale_continuous_breaks():
    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    breaks = [2, 4, 6, 8, 10]

    # Array breaks should not trip up the conditional checks
    s1 = scale_x_continuous(breaks=breaks, limits=(1, 10))
    s2 = scale_x_continuous(breaks=np.array(breaks), limits=(1, 10))
    s1.train(x)
    s2.train(x)
    assert list(s1.get_breaks()) == list(s2.get_breaks())


def test_no_scale_continuous_breaks():
    x = list(range(2, 12))
    sc1 = scale_x_continuous(breaks=False)
    sc1.train(x)
    assert not sc1.get_breaks()
    assert not sc1.get_labels()

    sc2 = scale_x_continuous(breaks=None)
    sc2.train(x)
    assert not sc2.get_breaks()
    assert not sc2.get_labels()


def test_scale_discrete_breaks():
    x = pd.Series(list("abccdefe"))
    breaks = list("ace")

    sc1 = scale_x_discrete(breaks=breaks)
    sc2 = scale_x_discrete(limits=breaks)
    sc1.train(x)
    sc2.train(x)
    assert list(sc1.get_breaks()) == list(sc2.get_breaks())


def test_no_scale_discrete_breaks():
    x = pd.Series(list("abccdefe"))

    # Array breaks should not trip up the conditional checks
    sc1 = scale_x_discrete(breaks=False)
    sc1.train(x)
    assert not sc1.get_breaks()
    assert not sc1.get_labels()

    sc2 = scale_x_discrete(breaks=None)
    sc2.train(x)
    assert not sc2.get_breaks()
    assert not sc2.get_labels()


def test_scale_without_a_mapping():
    df = pd.DataFrame(
        {
            "x": [1, 2, 3],
        }
    )
    p = (
        ggplot(df, aes("x", "x"))
        + geom_point()
        + scale_color.scale_color_continuous()
    )
    with pytest.warns(PlotnineWarning):
        p.draw_test()


def test_scale_discrete_mapping_nulls():
    a = np.array([1, 2, 3], dtype=object)

    sc = _scale_manual([1, 2, 3, 4, 5])
    sc.train(a)
    res = sc.map([1, 2, 3])
    expected = np.array([1, 2, 3])
    npt.assert_array_equal(res, expected)

    sc = _scale_manual([1, None, 3, 4, 5])
    sc.train(a)
    res = sc.map([1, 2, 3])
    expected = np.array([1, np.nan, 3])
    assert res[0] == expected[0]
    assert all(np.isnan([res[1], expected[1]]))
    assert res[2] == expected[2]


def test_scale_continuous_mapping_nulls():
    # Handling of nans
    sc = scale_color.scale_fill_gradient("yellow", "blue", na_value="green")
    sc.train([1, 10])
    res = sc.map([1, 5, np.nan, 10])
    assert res[2] == "green"


def test_multiple_aesthetics():
    df = pd.DataFrame({"x": [1, 2, 3], "y": [-1, -2, -3]})
    p = (
        ggplot(df, aes("x", "x", color="factor(x)", fill="factor(y)"))
        + geom_point(size=9, stroke=2)
        + scale_color.scale_color_brewer(
            type="qual", palette=1, aesthetics=["fill", "color"]
        )
    )
    assert p == "multiple_aesthetics"


def test_missing_manual_dict_aesthetic():
    df = pd.DataFrame(
        {"x": range(15), "y": range(15), "c": np.repeat(list("ABC"), 5)}
    )

    values = {"A": "red", "B": "violet", "D": "blue"}

    p = (
        ggplot(df, aes("x", "y", color="c"))
        + geom_point(size=3)
        + scale_manual.scale_color_manual(values)
    )
    assert p == "missing_manual_dict_aesthetic"


def test_missing_data_discrete_scale():
    df = pd.DataFrame({"a": [1, 2, 3], "b": ["a", "b", np.nan]})

    p = ggplot(df, aes("a", "a")) + geom_point(
        aes(fill="b"), stroke=0, size=10
    )
    assert p == "missing_data_discrete_scale"


def test_missing_data_discrete_position_scale():
    df = pd.DataFrame({"a": [1, 2, 3], "b": ["a", "b", None]})

    p = ggplot(df, aes("a", "b")) + geom_point(
        aes(fill="b"), stroke=0, size=10
    )

    with pytest.warns(PlotnineWarning):
        assert p == "missing_data_discrete_position_scale"


df = pd.DataFrame(
    {"x": range(4), "y": range(4), "w": list("wxyz"), "z": list("abcd")}
)

# Order of legend
# The precedence is driven by
# 1. The order in which the scales are added
# 2. Order of aesthetics in the local (geom) aes calls
# 2. Order of aesthetics in the global (ggplot)  aes calls


def test_legend_ordering_global_aethetics_1():
    # 1. color
    # 2. shape
    p = ggplot(df) + aes("x", "y", color="w", shape="z") + geom_point(size=5)

    assert p == "legend_ordering_global_aesthetics_1"


def test_legend_ordering_global_aesthetics_2():
    # 1. shape
    # 2. color
    p = ggplot(df) + aes("x", "y", shape="z", color="w") + geom_point(size=5)

    assert p == "legend_ordering_global_aesthetics_2"


def test_legend_ordering_local_aethetics_1():
    # 1. color
    # 2. shape
    p = (
        ggplot(df)
        + aes("x", "y")
        + geom_point(aes(color="w", shape="z"), size=5)
    )

    assert p == "legend_ordering_local_aesthetics_1"


def test_legend_ordering_local_aethetics_2():
    # 1. shape
    # 2. color
    p = (
        ggplot(df)
        + aes("x", "y")
        + geom_point(aes(shape="z", color="w"), size=5)
    )

    assert p == "legend_ordering_local_aesthetics_2"


def test_legend_ordering_mixed_scope_aesthetics():
    # The local(geom) aesthetics come first.
    # 1. color
    # 2. shape
    p = (
        ggplot(df)
        + aes("x", "y", shape="z")
        + geom_point(aes(color="w"), size=5)
    )

    assert p == "legend_ordering_mixed_scope_aesthetics"


def test_legend_ordering_added_scales():
    # The first added scale comes first
    # 1. color
    # 2. shape
    p = (
        ggplot(df)
        + aes("x", "y")
        + geom_point(aes(shape="z", color="w"), size=5)
        + scale_color.scale_color_discrete()
    )

    assert p == "legend_ordering_added_scales"


def test_breaks_and_labels_outside_of_limits():
    df = pd.DataFrame({"x": range(5, 11), "y": range(5, 11)})
    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + scale_x_continuous(
            limits=[7, 9.5],
            breaks=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
            labels=[
                "one",
                "two",
                "three",
                "four",
                "five",
                "six",
                "seven",
                "eight",
                "nine",
                "ten",
                "eleven",
            ],
        )
    )
    with pytest.warns(PlotnineWarning):
        assert p == "breaks_and_labels_outside_of_limits"


def test_changing_scale_transform():
    # No warning
    with warnings.catch_warnings(record=True) as record:
        scale_x_continuous(trans="reverse")
        scale_xy.scale_x_reverse(trans="reverse")
        scale_xy.scale_x_log10(trans="log10")
        assert not record, "Issued an unexpected warning"

    # Warnings
    with pytest.warns(PlotnineWarning):
        scale_xy.scale_x_reverse(trans="log10")

    with pytest.warns(PlotnineWarning):
        scale_xy.scale_x_datetime(trans="identity")

    s = scale_xy.scale_x_reverse()
    with pytest.warns(PlotnineWarning):
        s.trans = "log10"


def test_datetime_scale_limits():
    n = 6

    df = pd.DataFrame(
        {
            "x": [datetime(x, 1, 1) for x in range(2000, 2000 + n)],
            "y": range(n),
        }
    )

    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + lims(x=[datetime(2000, 1, 1), datetime(2007, 1, 1)])
        + theme(axis_text_x=element_text(angle=45))
    )

    assert p == "datetime_scale_limits"


def test_ordinal_scale():
    df = pd.DataFrame(
        {"x": pd.Categorical(list("abcd"), ordered=True), "y": [1, 2, 3, 4]}
    )

    p = ggplot(df) + aes("x", "y", color="-y", fill="x") + geom_col(size=4)

    assert p == "ordinal_scale"


def test_layer_with_only_infs():
    df = pd.DataFrame({"x": ["a", "b"]})
    p = (
        ggplot(df, aes("x", "x"))
        + annotate(
            "rect",
            xmin=-np.inf,
            xmax=np.inf,
            ymin=-np.inf,
            ymax=np.inf,
            fill="black",
            alpha=0.25,
        )
        + geom_point(color="red", size=3)
    )
    p = p.build_test()
    assert isinstance(p.scales.get_scales("x"), scale_x_discrete)


def test_discrete_scale_exceeding_maximum_number_of_values():
    df = pd.DataFrame(
        {
            # not that it's the second c that triggered a bug
            # in scale_discrete.map
            "x": pd.Categorical(["c", "a", "c", "b", "c"]),
            "y": [0, 1, 2, 2, 3],
        }
    )
    p = (
        ggplot(df, aes("x", "y", color="x", shape="x"))
        + geom_point()
        + scale_color_manual(["red", "blue"])
    )
    with pytest.warns(PlotnineWarning):
        p.draw_test()


def test_discrete_scale_for_empty_layer():
    # Ref: https://github.com/has2k1/plotnine/issues/647
    df1 = pd.DataFrame({"x": list("abc"), "y": [1, 2, 3], "g": list("AAA")})

    df2 = pd.DataFrame({"x": list("abc"), "y": [4, 5, 6], "g": list("AAB")})

    p = (
        ggplot(aes("x", "y"))
        + geom_point(df1)
        + geom_point(df2)
        + facet_wrap("g", scales="free_x")
    )

    p.draw_test()



from __future__ import annotations

import typing

from .coord_cartesian import coord_cartesian

if typing.TYPE_CHECKING:
    from typing import Optional

    from plotnine.iapi import panel_view
    from plotnine.typing import TupleFloat2


class coord_fixed(coord_cartesian):
    """
    Cartesian coordinates with fixed relationship between x and y scales

    Parameters
    ----------
    ratio : float
        Desired aspect_ratio (:math:`y/x`) of the panel(s).
        Default is 1.
    xlim : None | (float, float)
        Limits for x axis. If None, then they are
        automatically computed.
    ylim : None | (float, float)
        Limits for y axis. If None, then they are
        automatically computed.
    expand : bool
        If `True`, expand the coordinate axes by
        some factor. If `False`, use the limits
        from the data.

    Notes
    -----
    To specify aspect ratio of the visual size for the axes use the
    :class:`~plotnine.themes.themeable.aspect_ratio` themeable::

        ggplot(data, aes('x', 'y')) + theme(aspect_ratio=0.5)

    When changing the `aspect_ratio` in either way, the `width` of the
    panel remains constant (as derived from the
    :class:`plotnine.themes.themeable.figure_size` themeable) and the
    `height` is altered to achieve desired ratio.
    """

    ratio: float

    def __init__(
        self,
        ratio: float = 1,
        xlim: Optional[TupleFloat2] = None,
        ylim: Optional[TupleFloat2] = None,
        expand: bool = True,
    ):
        super().__init__(xlim=xlim, ylim=ylim, expand=expand)
        self.ratio = ratio

    def aspect(self, panel_params: panel_view) -> float | None:
        x = panel_params.x.range
        y = panel_params.y.range
        return (y[1] - y[0]) / (x[1] - x[0]) * self.ratio


coord_equal = coord_fixed



import itertools
import types

import numpy as np
import pandas as pd

from ..doctools import document
from ..mapping.evaluation import after_stat
from ..utils import is_scalar_or_string
from .binning import fuzzybreaks
from .stat import stat


@document
class stat_bin_2d(stat):
    """
    2 Dimensional bin counts

    {usage}

    Parameters
    ----------
    {common_parameters}
    bins : int, optional (default: 30)
        Number of bins. Overridden by binwidth.
    breaks : array-like(s), optional (default: None)
        Bin boundaries. This supercedes the ``binwidth``, ``bins``,
        ``center`` and ``boundary``. It can be an array_like or
        a list of two array_likes to provide distinct breaks for
        the ``x`` and ``y`` axes.
    binwidth : float, optional (default: None)
        The width of the bins. The default is to use bins bins that
        cover the range of the data. You should always override this
        value, exploring multiple widths to find the best to illustrate
        the stories in your data.
    drop : bool, optional (default: False)
        If :py:`True`, removes all cells with zero counts.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'xmin'    # x lower bound for the bin
        'xmax'    # x upper bound for the bin
        'ymin'    # y lower bound for the bin
        'ymax'    # y upper bound for the bin
        'count'   # number of points in bin
        'density' # density of points in bin, scaled to integrate to 1

    """
    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "rect",
        "position": "identity",
        "na_rm": False,
        "bins": 30,
        "breaks": None,
        "binwidth": None,
        "drop": True,
    }
    DEFAULT_AES = {"fill": after_stat("count"), "weight": None}
    CREATES = {"xmin", "xmax", "ymin", "ymax", "count", "density"}

    def setup_params(self, data):
        params = self.params.copy()
        params["bins"] = dual_param(params["bins"])
        params["breaks"] = dual_param(params["breaks"])
        params["binwidth"] = dual_param(params["binwidth"])
        return params

    @classmethod
    def compute_group(cls, data, scales, **params):
        bins = params["bins"]
        breaks = params["breaks"]
        binwidth = params["binwidth"]
        drop = params["drop"]
        weight = data.get("weight")

        if weight is None:
            weight = np.ones(len(data["x"]))

        # The bins will be over the dimension(full size) of the
        # trained x and y scales
        range_x = scales.x.dimension()
        range_y = scales.y.dimension()

        # Trick pd.cut into creating cuts over the range of
        # the scale
        x = np.append(data["x"], range_x)
        y = np.append(data["y"], range_y)

        # create the cutting parameters
        xbreaks = fuzzybreaks(
            scales.x, breaks=breaks.x, binwidth=binwidth.x, bins=bins.x
        )
        ybreaks = fuzzybreaks(
            scales.y, breaks.y, binwidth=binwidth.y, bins=bins.y
        )

        xbins = pd.cut(
            x,
            bins=xbreaks,  # pyright: ignore
            labels=False,  # pyright: ignore
            right=True,
        )
        ybins = pd.cut(
            y,
            bins=ybreaks,  # pyright: ignore
            labels=False,  # pyright: ignore
            right=True,
        )

        # Remove the spurious points
        xbins = xbins[:-2]
        ybins = ybins[:-2]

        # Because we are graphing, we want to see equal breaks
        # The original breaks have an extra room to the left
        ybreaks[0] -= np.diff(np.diff(ybreaks))[0]
        xbreaks[0] -= np.diff(np.diff(xbreaks))[0]

        df = pd.DataFrame(
            {
                "xbins": xbins,
                "ybins": ybins,
                "weight": weight,
            }
        )
        table = df.pivot_table(
            "weight", index=["xbins", "ybins"], aggfunc=np.sum
        )["weight"]

        # create rectangles
        rects = []
        keys = itertools.product(
            range(len(ybreaks) - 1), range(len(xbreaks) - 1)
        )
        for j, i in keys:
            try:
                cval = table[(i, j)]
            except KeyError:
                if drop:
                    continue
                cval = 0
            # xmin, xmax, ymin, ymax, count
            row = [
                xbreaks[i],
                xbreaks[i + 1],
                ybreaks[j],
                ybreaks[j + 1],
                cval,
            ]
            rects.append(row)

        new_data = pd.DataFrame(
            rects, columns=["xmin", "xmax", "ymin", "ymax", "count"]
        )
        new_data["density"] = new_data["count"] / new_data["count"].sum()
        return new_data


stat_bin2d = stat_bin_2d


def dual_param(value):
    """
    Return duplicate of parameter value

    Used to apply same value to x & y axes if only one
    value is given.
    """
    if is_scalar_or_string(value):
        return types.SimpleNamespace(x=value, y=value)

    if hasattr(value, "x") and hasattr(value, "y"):
        return value

    if len(value) == 2:
        return types.SimpleNamespace(x=value[0], y=value[1])
    else:
        return types.SimpleNamespace(x=value, y=value)



from __future__ import annotations

import typing

from .facet import facet, layout_null

if typing.TYPE_CHECKING:
    import pandas as pd


class facet_null(facet):
    """
    A single Panel

    Parameters
    ----------
    shrink : bool
        Whether to shrink the scales to the output of the
        statistics instead of the raw data. Default is ``True``.
    """

    def __init__(self, shrink: bool = True):
        facet.__init__(self, shrink=shrink)
        self.nrow = 1
        self.ncol = 1

    def map(self, data: pd.DataFrame, layout: pd.DataFrame) -> pd.DataFrame:
        data["PANEL"] = 1
        return data

    def compute_layout(
        self,
        data: list[pd.DataFrame],
    ) -> pd.DataFrame:
        return layout_null()



from __future__ import annotations

import typing
from warnings import warn

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import alias
from .scale import scale_continuous, scale_datetime, scale_discrete

if typing.TYPE_CHECKING:
    from typing import Literal


# Discrete color scales #
# Note: plotnine operates in the hcl space
@document
class scale_color_hue(scale_discrete):
    """
    Qualitative color scale with evenly spaced hues

    Parameters
    ----------
    h : float
        first hue. Must be in the range [0, 1]
        Default is ``0.01``
    l : float
        lightness. Must be in the range [0, 1]
        Default is ``0.6``
    s : float
        saturation. Must be in the range [0, 1]
        Default is ``0.65``
    colorspace : str in ``['hls', 'husl']``
        Color space to use.
        `hls <https://en.wikipedia.org/wiki/HSL_and_HSV>`_
        `husl <http://www.husl-colors.org/>`_
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'#7F7F7F'``
    """

    _aesthetics = ["color"]
    na_value = "#7F7F7F"

    def __init__(self, h=0.01, l=0.6, s=0.65, color_space="hls", **kwargs):
        from mizani.palettes import hue_pal

        self.palette = hue_pal(h, l, s, color_space=color_space)
        scale_discrete.__init__(self, **kwargs)


@document
class scale_fill_hue(scale_color_hue):
    """
    Qualitative color scale with evenly spaced hues

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_brewer(scale_discrete):
    """
    Sequential, diverging and qualitative discrete color scales

    See `colorbrewer.org <http://colorbrewer2.org/>`_

    Parameters
    ----------
    type : str in ``['seq', 'div', 'qual']``
        Type of data. Sequential, diverging or qualitative
    palette : int | str
         If a string, will use that named palette.
         If a number, will index into the list of palettes
         of appropriate type. Default is 1
    direction: int in ``[-1, 1]``
         Sets the order of colors in the scale. If 1,
         the default, colors are as output by
         mizani.palettes.brewer_pal(). If -1,
         the order of colors is reversed.
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``
    """

    _aesthetics = ["color"]
    na_value = "#7F7F7F"

    def __init__(
        self, type="seq", palette=1, direction: Literal[1, -1] = 1, **kwargs
    ):
        from mizani.palettes import brewer_pal

        self.palette = brewer_pal(type, palette, direction=direction)
        scale_discrete.__init__(self, **kwargs)


@document
class scale_fill_brewer(scale_color_brewer):
    """
    Sequential, diverging and qualitative color scales

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_grey(scale_discrete):
    """
    Sequential grey color scale.

    Parameters
    ----------
    start : float
        grey value at low end of palette.
        Default is 0.2
    end : float
        grey value at high end of palette
        Default is 0.8
    {superclass_parameters}
    """

    _aesthetics = ["color"]

    def __init__(self, start=0.2, end=0.8, **kwargs):
        from mizani.palettes import grey_pal

        # TODO: fix types in mizani
        self.palette = grey_pal(start, end)  # pyright: ignore
        scale_discrete.__init__(self, **kwargs)


@document
class scale_fill_grey(scale_color_grey):
    """
    Sequential grey color scale.

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


# Continuous color scales #


@document
class scale_color_gradient(scale_continuous):
    """
    Create a 2 point color gradient

    Parameters
    ----------
    low : str
        low color
    high : str
        high color
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``

    See Also
    --------
    :class:`.scale_color_gradient2`
    :class:`.scale_color_gradientn`
    """

    _aesthetics = ["color"]
    guide = "colorbar"
    na_value = "#7F7F7F"

    def __init__(self, low="#132B43", high="#56B1F7", **kwargs):
        """
        Create colormap that will be used by the palette
        """
        from mizani.palettes import gradient_n_pal

        # TODO: fix types in mizani
        self.palette = gradient_n_pal(
            [low, high], name="gradient"
        )  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_fill_gradient(scale_color_gradient):
    """
    Create a 2 point color gradient

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_desaturate(scale_continuous):
    """
    Create a desaturated color gradient

    Parameters
    ----------
    color : str, optional (Default: 'red')
        Color to desaturate
    prop : float, optional (Default: 0)
        Saturation channel of color will be multiplied by
        this value.
    reverse : bool, optional (Default: False)
        Whether to go from color to desaturated color
        or desaturated color to color.
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``
    """

    _aesthetics = ["color"]
    guide = "colorbar"
    na_value = "#7F7F7F"

    def __init__(self, color="red", prop=0, reverse=False, **kwargs):
        from mizani.palettes import desaturate_pal

        # TODO: fix types in mizani
        self.palette = desaturate_pal(color, prop, reverse)  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_fill_desaturate(scale_color_desaturate):
    """
    Create a desaturated color gradient

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_gradient2(scale_continuous):
    """
    Create a 3 point diverging color gradient

    Parameters
    ----------
    low : str, optional
        low color
    mid : str, optional
        mid point color
    high : str, optional
        high color
    midpoint : float, optional (Default: 0)
        Mid point of the input data range.
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``

    See Also
    --------
    :class:`.scale_color_gradient`
    :class:`.scale_color_gradientn`
    """

    _aesthetics = ["color"]
    guide = "colorbar"
    na_value = "#7F7F7F"

    def __init__(
        self,
        low="#832424",
        mid="#FFFFFF",
        high="#3A3A98",
        midpoint=0,
        **kwargs,
    ):
        from mizani.bounds import rescale_mid
        from mizani.palettes import gradient_n_pal

        # All rescale functions should have the same signature
        def _rescale_mid(*args, **kwargs):
            return rescale_mid(*args, mid=midpoint, **kwargs)

        kwargs["rescaler"] = _rescale_mid
        # TODO: fix types in mizani
        self.palette = gradient_n_pal(
            [low, mid, high], name="gradient2"
        )  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_fill_gradient2(scale_color_gradient2):
    """
    Create a 3 point diverging color gradient

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_gradientn(scale_continuous):
    """
    Create a n color gradient

    Parameters
    ----------
    colors : list
        list of colors
    values : list, optional
        list of points in the range [0, 1] at which to
        place each color. Must be the same size as
        `colors`. Default to evenly space the colors
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``

    See Also
    --------
    :class:`.scale_color_gradient`
    :class:`.scale_color_gradientn`
    """

    _aesthetics = ["color"]
    guide = "colorbar"
    na_value = "#7F7F7F"

    def __init__(self, colors, values=None, **kwargs):
        from mizani.palettes import gradient_n_pal

        # TODO: fix types in mizani
        self.palette = gradient_n_pal(
            colors, values, "gradientn"
        )  # pyright: ignore
        scale_continuous.__init__(self, **kwargs)


@document
class scale_fill_gradientn(scale_color_gradientn):
    """
    Create a n color gradient

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_distiller(scale_color_gradientn):
    """
    Sequential and diverging continuous color scales

    This is a convinience scale around :class:`.scale_color_gradientn`
    with colors from `colorbrewer.org <http://colorbrewer2.org/>`_.
    It smoothly interpolates 7 colors from a brewer palette to create
    a continuous palette.

    Parameters
    ----------
    type : str in ``['seq', 'div']``
        Type of data. Sequential, diverging or qualitative
    palette : int | str
         If a string, will use that named palette.
         If a number, will index into the list of palettes
         of appropriate type. Default is 1
    values : list, optional
        list of points in the range [0, 1] at which to
        place each color. Must be the same size as
        `colors`. Default to evenly space the colors
    direction: int in ``[-1, 1]``
        Sets the order of colors in the scale. If 1
        colors are as output by mizani.palettes.brewer_pal().
        If -1, the default, the order of colors is reversed.
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``
    """

    _aesthetics = ["color"]
    guide = "colorbar"
    na_value = "#7F7F7F"

    def __init__(
        self,
        type="seq",
        palette=1,
        values=None,
        direction: Literal[1, -1] = -1,
        **kwargs,
    ):
        """
        Create colormap that will be used by the palette
        """
        from mizani.palettes import brewer_pal

        if type.lower() in ("qual", "qualitative"):
            warn(
                "Using a discrete color palette in a continuous scale."
                "Consider using type = 'seq' or type = 'div' instead",
                PlotnineWarning,
            )

        # Grab 7 colors from brewer and create a gradient palette
        # An odd number matches the midpoint of the palette to that
        # of the data
        colors = brewer_pal(type, palette, direction=direction)(7)
        scale_color_gradientn.__init__(self, colors, values, **kwargs)


@document
class scale_fill_distiller(scale_color_distiller):
    """
    Sequential, diverging continuous color scales

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


# matplotlib colormaps
@document
class scale_color_cmap(scale_continuous):
    """
    Create color scales using Matplotlib colormaps

    Parameters
    ----------
    cmap_name : str
        A standard Matplotlib colormap name. The default is
        `viridis`. For the list of names checkout the output
        of ``matplotlib.cm.cmap_d.keys()`` or see the
        `documentation <http://matplotlib.org/users/colormaps.html>`_.
    lut : None | int
        This is the number of entries desired in the
        lookup table. Default is `None`, leave it up
        Matplotlib.
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``

    See Also
    --------
    :mod:`matplotlib.cm`
    :mod:`matplotlib.colors`
    """

    _aesthetics = ["color"]
    guide = "colorbar"
    na_value = "#7F7F7F"

    def __init__(self, cmap_name="viridis", lut=None, **kwargs):
        from mizani.palettes import cmap_pal

        # TODO: fix types in mizani
        self.palette = cmap_pal(cmap_name, lut)  # pyright: ignore
        super().__init__(**kwargs)


@document
class scale_fill_cmap(scale_color_cmap):
    """
    Create color scales using Matplotlib colormaps

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_cmap_d(scale_discrete):
    """
    A discrete color scales using Matplotlib colormaps

    Parameters
    ----------
    name : str
        A standard Matplotlib colormap name. It must be of type
        :class:`matplotlib.colors.ListedColormap`.
        . The default is `viridis`. For the list of names checkout
        the output of ``matplotlib.cm.cmap_d.keys()`` or see the
        `documentation <http://matplotlib.org/users/colormaps.html>`_.
    lut : None | int
        This is the number of entries desired in the
        lookup table. Default is `None`, leave it up
        Matplotlib.
    {superclass_parameters}
    na_value : str
        Color of missing values. Default is ``'None'``

    See Also
    --------
    :mod:`matplotlib.cm`
    :mod:`matplotlib.colors`
    """

    _aesthetics = ["color"]
    na_value = "#7F7F7F"

    def __init__(self, name="viridis", lut=None, **kwargs):
        from mizani.palettes import cmap_d_pal

        self.palette = cmap_d_pal(name, lut)
        super().__init__(**kwargs)


@document
class scale_fill_cmap_d(scale_color_cmap_d):
    """
    Create color scales using Matplotlib colormaps

    Parameters
    ----------
    {superclass_parameters}
    """

    _aesthetics = ["fill"]


@document
class scale_color_datetime(scale_datetime, scale_color_cmap):
    """
    Datetime color scale

    Parameters
    ----------
    {superclass_parameters}
    """


@document
class scale_fill_datetime(scale_datetime, scale_fill_cmap):
    """
    Datetime fill scale

    Parameters
    ----------
    {superclass_parameters}
    """


# Default scales
alias("scale_color_discrete", scale_color_hue)
alias("scale_color_continuous", scale_color_cmap)
alias("scale_color_ordinal", scale_color_cmap_d)
alias("scale_fill_discrete", scale_fill_hue)
alias("scale_fill_continuous", scale_fill_cmap)
alias("scale_fill_ordinal", scale_fill_cmap_d)

# American to British spelling
alias("scale_colour_hue", scale_color_hue)
alias("scale_color_gray", scale_color_grey)
alias("scale_colour_grey", scale_color_grey)
alias("scale_colour_gray", scale_color_grey)
alias("scale_fill_gray", scale_fill_grey)
alias("scale_colour_brewer", scale_color_brewer)
alias("scale_colour_desaturate", scale_color_desaturate)
alias("scale_colour_gradient", scale_color_gradient)
alias("scale_colour_gradient2", scale_color_gradient2)
alias("scale_colour_gradientn", scale_color_gradientn)
alias("scale_colour_discrete", scale_color_hue)
alias("scale_colour_continuous", scale_color_cmap)
alias("scale_colour_distiller", scale_color_distiller)
alias("scale_colour_cmap", scale_color_cmap)
alias("scale_colour_cmap_d", scale_color_cmap_d)
alias("scale_colour_datetime", scale_color_datetime)
alias("scale_colour_ordinal", scale_color_cmap_d)



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import SIZE_FACTOR, copy_missing_columns, resolution, to_rgba
from .geom import geom
from .geom_polygon import geom_polygon
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer


@document
class geom_crossbar(geom):
    """
    Vertical interval represented by a crossbar

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float or None, optional (default: 0.5)
        Box width. If :py:`None`, the width is set to
        `90%` of the resolution of the data.
    fatten : float, optional (default: 2)
        A multiplicative factor used to increase the size of the
        middle bar across the box.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "fill": None,
        "linetype": "solid",
        "size": 0.5,
    }
    REQUIRED_AES = {"x", "y", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "width": 0.5,
        "fatten": 2,
    }

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if "width" not in data:
            if self.params["width"]:
                data["width"] = self.params["width"]
            else:
                data["width"] = resolution(data["x"], False) * 0.9

        data["xmin"] = data["x"] - data["width"] / 2
        data["xmax"] = data["x"] + data["width"] / 2
        del data["width"]
        return data

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        y = data["y"]
        xmin = data["xmin"]
        xmax = data["xmax"]
        ymin = data["ymin"]
        ymax = data["ymax"]
        group = data["group"]

        # From violin
        notchwidth = typing.cast(float, params.get("notchwidth"))
        # ynotchupper = data.get('ynotchupper')
        # ynotchlower = data.get('ynotchlower')

        def flat(*args: pd.Series[Any]) -> npt.NDArray[Any]:
            """Flatten list-likes"""
            return np.hstack(args)

        middle = pd.DataFrame(
            {"x": xmin, "y": y, "xend": xmax, "yend": y, "group": group}
        )
        copy_missing_columns(middle, data)
        middle["alpha"] = 1
        middle["size"] *= params["fatten"]

        has_notch = "ynotchupper" in data and "ynotchlower" in data
        if has_notch:  # 10 points + 1 closing
            ynotchupper = data["ynotchupper"]
            ynotchlower = data["ynotchlower"]

            if any(ynotchlower < ymin) or any(ynotchupper > ymax):
                warn(
                    "Notch went outside the hinges. "
                    "Try setting notch=False.",
                    PlotnineWarning,
                )

            notchindent = (1 - notchwidth) * (xmax - xmin) / 2

            middle["x"] += notchindent
            middle["xend"] -= notchindent
            box = pd.DataFrame(
                {
                    "x": flat(
                        xmin,
                        xmin,
                        xmin + notchindent,
                        xmin,
                        xmin,
                        xmax,
                        xmax,
                        xmax - notchindent,
                        xmax,
                        xmax,
                        xmin,
                    ),
                    "y": flat(
                        ymax,
                        ynotchupper,
                        y,
                        ynotchlower,
                        ymin,
                        ymin,
                        ynotchlower,
                        y,
                        ynotchupper,
                        ymax,
                        ymax,
                    ),
                    "group": np.tile(np.arange(1, len(group) + 1), 11),
                }
            )
        else:
            # No notch, 4 points + 1 closing
            box = pd.DataFrame(
                {
                    "x": flat(xmin, xmin, xmax, xmax, xmin),
                    "y": flat(ymax, ymax, ymax, ymin, ymin),
                    "group": np.tile(np.arange(1, len(group) + 1), 5),
                }
            )

        copy_missing_columns(box, data)
        geom_polygon.draw_group(box, panel_params, coord, ax, **params)
        geom_segment.draw_group(middle, panel_params, coord, ax, **params)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a rectangle with a horizontal strike in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D
        from matplotlib.patches import Rectangle

        data["size"] *= SIZE_FACTOR

        # background
        facecolor = to_rgba(data["fill"], data["alpha"])
        if facecolor is None:
            facecolor = "none"

        bg = Rectangle(
            (da.width * 0.125, da.height * 0.25),
            width=da.width * 0.75,
            height=da.height * 0.5,
            linewidth=data["size"],
            facecolor=facecolor,
            edgecolor=data["color"],
            linestyle=data["linetype"],
            capstyle="projecting",
            antialiased=False,
        )
        da.add_artist(bg)

        strike = Line2D(
            [da.width * 0.125, da.width * 0.875],
            [da.height * 0.5, da.height * 0.5],
            linestyle=data["linetype"],
            linewidth=data["size"],
            color=data["color"],
        )
        da.add_artist(strike)
        return da



from __future__ import annotations

import typing

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..mapping.evaluation import after_stat
from ..scales.scale import scale_continuous
from .stat import stat

if typing.TYPE_CHECKING:
    from typing import Callable

    from plotnine.typing import FloatArrayLike


@document
class stat_function(stat):
    """
    Superimpose a function onto a plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    fun : function
        Function to evaluate.
    n : int, optional (default: 101)
        Number of points at which to evaluate the function.
    xlim : tuple (default: None)
        ``x`` limits for the range. The default depends on
        the ``x`` aesthetic. There is not an ``x`` aesthetic
        then the ``xlim`` must be provided.
    args : tuple or dict (default: None)
        Arguments to pass to ``fun``.
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'x'  # x points are which the function is evaluated
        'y'  # Points evaluated at x
    """

    DEFAULT_PARAMS = {
        "geom": "path",
        "position": "identity",
        "na_rm": False,
        "fun": None,
        "n": 101,
        "args": None,
        "xlim": None,
    }

    DEFAULT_AES = {"y": after_stat("y")}
    CREATES = {"y"}

    def __init__(self, mapping=None, data=None, **kwargs):
        if data is None:

            def _data_func(df: pd.DataFrame) -> pd.DataFrame:
                if df.empty:
                    df = pd.DataFrame({"group": [1]})
                return df

            data = _data_func

        super().__init__(mapping, data, **kwargs)

    def setup_params(self, data):
        if not callable(self.params["fun"]):
            raise PlotnineError(
                "stat_function requires parameter 'fun' to be "
                "a function or any other callable object"
            )
        return self.params

    @classmethod
    def compute_group(cls, data, scales, **params):
        fun: Callable[..., FloatArrayLike] = params["fun"]  # pyright: ignore
        n = params["n"]
        args = params["args"]
        xlim = params["xlim"]
        range_x = xlim or scales.x.dimension((0, 0))
        old_fun = fun

        if isinstance(args, (list, tuple)):

            def fun(x):
                return old_fun(x, *args)

        elif isinstance(args, dict):

            def fun(x):
                return old_fun(x, **args)

        elif args is not None:

            def fun(x):
                return old_fun(x, args)

        else:

            def fun(x):
                return old_fun(x)

        x = np.linspace(range_x[0], range_x[1], n)

        # continuous scale
        if isinstance(scales.x, scale_continuous):
            x = scales.x.trans.inverse(x)

        # We know these can handle array-likes
        if isinstance(old_fun, (np.ufunc, np.vectorize)):
            y = fun(x)
        else:
            y = [fun(val) for val in x]

        new_data = pd.DataFrame({"x": x, "y": y})
        return new_data



import numpy as np
import pandas as pd

from plotnine import aes, geom_point, geom_quantile, ggplot

n = 200  # Should not be too big, affects the test duration
random_state = np.random.RandomState(1234567890)
# points that diverge like a point flash-light
df = pd.DataFrame(
    {"x": np.arange(n), "y": np.arange(n) * (1 + random_state.rand(n))}
)


def test_lines():
    p = (
        ggplot(df, aes(x="x", y="y"))
        + geom_point(alpha=0.5)
        + geom_quantile(quantiles=[0.001, 0.5, 0.999], formula="y~x", size=2)
    )

    # Two (.001, .999) quantile lines should bound the points
    # from below and from above, and the .5 line should go
    # through middle (approximately).
    assert p == "lines"



from ..doctools import document
from .geom_point import geom_point


@document
class geom_count(geom_point):
    """
    Plot overlapping points

    This is a variant :class:`geom_point` that counts the number
    of observations at each location, then maps the count to point
    area. It useful when you have discrete data and overplotting.

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {"stat": "sum", "position": "identity", "na_rm": False}



from ..doctools import document
from .geom_bar import geom_bar


@document
class geom_col(geom_bar):
    """
    Bar plot with base on the x-axis

    This is an alternate version of :class:`geom_bar` that maps
    the height of bars to an existing variable in your data. If
    you want the height of the bar to represent a count of cases,
    use :class:`geom_bar`.

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float, (default: None)
        Bar width. If :py:`None`, the width is set to
        `90%` of the resolution of the data.

    See Also
    --------
    plotnine.geoms.geom_bar
    """

    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"xmin", "xmax", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "stack",
        "na_rm": False,
        "width": None,
    }



from __future__ import annotations

import typing
from copy import copy, deepcopy
from typing import overload

from ..exceptions import PlotnineError
from ..options import get_option, set_option
from .themeable import Themeables, themeable

if typing.TYPE_CHECKING:
    from typing import Any, Type

    from plotnine.typing import Axes, Figure, Ggplot

# All complete themes are initiated with these rcparams. They
# can be overridden.
default_rcparams = {
    "axes.axisbelow": "True",
    "font.sans-serif": [
        "Helvetica",
        "DejaVu Sans",  # MPL ships with this one
        "Avant Garde",
        "Computer Modern Sans serif",
        "Arial",
    ],
    "font.serif": [
        "Times",
        "Palatino",
        "New Century Schoolbook",
        "Bookman",
        "Computer Modern Roman",
        "Times New Roman",
    ],
    "lines.antialiased": "True",
    "patch.antialiased": "True",
    "timezone": "UTC",
}


class theme:
    """
    Base class for themes

    In general, only complete themes should subclass this class.

    Parameters
    ----------
    complete : bool
        Themes that are complete will override any existing themes.
        themes that are not complete (ie. partial) will add to or
        override specific elements of the current theme. e.g::

            theme_gray() + theme_xkcd()

        will be completely determined by :class:`theme_xkcd`, but::

            theme_gray() + theme(axis_text_x=element_text(angle=45))

        will only modify the x-axis text.

    kwargs: dict
        kwargs are :ref:`themeables <themeables>`. The themeables are
        elements that are subclasses of `themeable`. Many themeables
        are defined using theme elements i.e

            - :class:`element_line`
            - :class:`element_rect`
            - :class:`element_text`

        These simply bind together all the aspects of a themeable
        that can be themed. See
        :class:`~plotnine.themes.themeable.themeable`.

    Notes
    -----
    When subclassing, make sure to call :python:`theme.__init__`.
    After which you can customise :python:`self._rcParams` within
    the ``__init__`` method of the new theme. The ``rcParams``
    should not be modified after that.
    """

    # This is set when the figure is created,
    # it is useful at legend drawing time and
    # when applying the theme.
    figure: Figure
    axs: list[Axes]
    complete: bool

    # Dictionary to collect matplotlib objects that will
    # be targeted for theming by the themeables
    # It is initialised in the plot context and removed at
    # the end of it.
    _targets: dict[str, Any]

    def __init__(
        self,
        complete=False,
        # Generate themeables keyword parameters with
        #
        #     from plotnine.themes.themeable import themeable
        #     for name in themeable.registry():
        #         print(f'{name}=None,')
        axis_title_x=None,
        axis_title_y=None,
        axis_title=None,
        legend_title=None,
        legend_text_legend=None,
        legend_text_colorbar=None,
        legend_text=None,
        plot_title=None,
        plot_subtitle=None,
        plot_caption=None,
        strip_text_x=None,
        strip_text_y=None,
        strip_text=None,
        title=None,
        axis_text_x=None,
        axis_text_y=None,
        axis_text=None,
        text=None,
        axis_line_x=None,
        axis_line_y=None,
        axis_line=None,
        axis_ticks_minor_x=None,
        axis_ticks_minor_y=None,
        axis_ticks_major_x=None,
        axis_ticks_major_y=None,
        axis_ticks_major=None,
        axis_ticks_minor=None,
        axis_ticks=None,
        panel_grid_major_x=None,
        panel_grid_major_y=None,
        panel_grid_minor_x=None,
        panel_grid_minor_y=None,
        panel_grid_major=None,
        panel_grid_minor=None,
        panel_grid=None,
        line=None,
        legend_key=None,
        legend_background=None,
        legend_box_background=None,
        panel_background=None,
        panel_border=None,
        plot_background=None,
        strip_background_x=None,
        strip_background_y=None,
        strip_background=None,
        rect=None,
        axis_ticks_length_major=None,
        axis_ticks_length_minor=None,
        axis_ticks_length=None,
        axis_ticks_pad_major=None,
        axis_ticks_pad_minor=None,
        axis_ticks_pad=None,
        axis_ticks_direction_x=None,
        axis_ticks_direction_y=None,
        axis_ticks_direction=None,
        panel_spacing_x=None,
        panel_spacing_y=None,
        panel_spacing=None,
        plot_margin=None,
        panel_ontop=None,
        aspect_ratio=None,
        dpi=None,
        figure_size=None,
        legend_box=None,
        legend_box_margin=None,
        legend_box_just=None,
        legend_direction=None,
        legend_key_width=None,
        legend_key_height=None,
        legend_key_size=None,
        legend_margin=None,
        legend_box_spacing=None,
        legend_spacing=None,
        legend_position=None,
        legend_title_align=None,
        legend_entry_spacing_x=None,
        legend_entry_spacing_y=None,
        legend_entry_spacing=None,
        strip_align_x=None,
        strip_align_y=None,
        strip_align=None,
        subplots_adjust=None,
        **kwargs,
    ):
        self.themeables = Themeables()
        self.complete = complete

        if complete:
            self._rcParams = deepcopy(default_rcparams)
        else:
            self._rcParams = {}

        # Themeables
        official_themeables = themeable.registry()
        locals_args = dict(locals())
        it = (
            (name, element)
            for name, element in locals_args.items()
            if element is not None and name in official_themeables
        )
        new = themeable.from_class_name

        for name, element in it:
            self.themeables[name] = new(name, element)

        # Unofficial themeables (for extensions)
        for name, element in kwargs.items():
            self.themeables[name] = new(name, element)

    def __eq__(self, other: Any) -> bool:
        """
        Test if themes are equal

        Mostly for testing purposes
        """
        # criteria for equality are
        # - Equal themeables
        # - Equal rcParams
        c1 = self.themeables == other.themeables
        c2 = self.rcParams == other.rcParams
        return c1 and c2

    def apply(self):
        """
        Apply this theme, then apply additional modifications in order.

        This method will be called once after plot has completed.
        Subclasses that override this method should make sure that the
        base class method is called.
        """
        for th in self.themeables.values():
            th.apply(self)

    def setup(self):
        """
        Setup theme & figure for before drawing

        1. The figure is modified with to the theme settings
           that it are required before drawing.
        2. Give contained objects of the theme/themeables a
           reference to the theme.

        This method will be called once with a figure object
        before any plotting has completed. Subclasses that
        override this method should make sure that the base
        class method is called.
        """
        from .elements import element_text

        for th in self.themeables.values():
            th.setup_figure(self.figure)
            if isinstance(th.theme_element, element_text):
                th.theme_element.setup(self)

    @property
    def rcParams(self):
        """
        Return rcParams dict for this theme.

        Notes
        -----
        Subclasses should not need to override this method method as long as
        self._rcParams is constructed properly.

        rcParams are used during plotting. Sometimes the same theme can be
        achieved by setting rcParams before plotting or a apply
        after plotting. The choice of how to implement it is is a matter of
        convenience in that case.

        There are certain things can only be themed after plotting. There
        may not be an rcParam to control the theme or the act of plotting
        may cause an entity to come into existence before it can be themed.

        """
        try:
            rcParams = deepcopy(self._rcParams)
        except NotImplementedError:
            # deepcopy raises an error for objects that are drived from or
            # composed of matplotlib.transform.TransformNode.
            # Not desirable, but probably requires upstream fix.
            # In particular, XKCD uses matplotlib.patheffects.withStrok
            rcParams = copy(self._rcParams)

        for th in self.themeables.values():
            rcParams.update(th.rcParams)
        return rcParams

    def add_theme(self, other: theme) -> theme:
        """
        Add themes together

        Subclasses should not override this method.

        This will be called when adding two instances of class 'theme'
        together.
        A complete theme will annihilate any previous themes. Partial themes
        can be added together and can be added to a complete theme.
        """
        if other.complete:
            return other

        self.themeables.update(deepcopy(other.themeables))
        return self

    def __add__(self, other: theme) -> theme:
        """
        Add other theme to this theme
        """
        if not isinstance(other, theme):
            msg = f"Adding theme failed. {other} is not a theme"
            raise PlotnineError(msg)
        self = deepcopy(self)
        return self.add_theme(other)

    @overload
    def __radd__(self, other: theme) -> theme:
        ...

    @overload
    def __radd__(self, other: Ggplot) -> Ggplot:
        ...

    def __radd__(self, other: theme | Ggplot) -> theme | Ggplot:
        """
        Add theme to ggplot object or to another theme

        This will be called in one of two ways::

             ggplot() + theme()
             theme1() + theme2()

        In both cases, `self` is the :class:`theme`
        on the right hand side.

        Subclasses should not override this method.
        """
        # ggplot() + theme, get theme
        # if hasattr(other, 'theme'):
        if not isinstance(other, theme):
            if self.complete:
                other.theme = self
            else:
                # If no theme has been added yet,
                # we modify the default theme
                other.theme = other.theme or theme_get()
                other.theme = other.theme.add_theme(self)
            return other
        # theme1 + theme2
        else:
            if self.complete:
                # e.g. other + theme_gray()
                return self
            else:
                # e.g. other + theme(...)
                return other.add_theme(self)

    def __iadd__(self, other: theme) -> theme:
        """
        Add theme to theme
        """
        return self.add_theme(other)

    def __deepcopy__(self, memo: dict) -> theme:
        """
        Deep copy without copying the figure
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        old = self.__dict__
        new = result.__dict__

        shallow = {"figure", "_targets"}
        for key, item in old.items():
            if key in shallow:
                new[key] = old[key]
                memo[id(new[key])] = new[key]
            else:
                new[key] = deepcopy(old[key], memo)

        return result


def theme_get() -> theme:
    """
    Return the default theme

    The default theme is the one set (using :func:`theme_set`) by
    the user. If none has been set, then :class:`theme_gray` is
    the default.
    """
    from .theme_gray import theme_gray

    _theme = get_option("current_theme")
    if isinstance(_theme, type):
        _theme = _theme()
    return _theme or theme_gray()


def theme_set(new: theme | Type[theme]) -> theme:
    """
    Change the current(default) theme

    Parameters
    ----------
    new : theme
        New default theme

    Returns
    -------
    out : theme
        Previous theme
    """
    if not isinstance(new, theme) and not issubclass(new, theme):
        raise PlotnineError("Expecting object to be a theme")

    out: theme = get_option("current_theme")
    set_option("current_theme", new)
    return out


def theme_update(**kwargs: themeable):
    """
    Modify elements of the current theme

    Parameters
    ----------
    kwargs : dict
        Theme elements
    """
    theme_set(theme_get() + theme(**kwargs))



from __future__ import annotations

import typing

from ..doctools import document
from ..exceptions import PlotnineError
from ..positions import position_jitter
from .geom_point import geom_point

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.typing import Aes, DataLike


@document
class geom_jitter(geom_point):
    """
    Scatter plot with points jittered to reduce overplotting

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float, optional
        Proportion to jitter in horizontal direction.
        The default value is that from
        :class:`~plotnine.positions.position_jitter`
    height : float, optional
        Proportion to jitter in vertical direction.
        The default value is that from
        :class:`~plotnine.positions.position_jitter`.
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.

    See Also
    --------
    plotnine.positions.position_jitter
    plotnine.geoms.geom_point
    """

    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "jitter",
        "na_rm": False,
        "width": None,
        "height": None,
        "random_state": None,
    }

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        if {"width", "height", "random_state"} & set(kwargs):
            if "position" in kwargs:
                raise PlotnineError(
                    "Specify either 'position' or "
                    "'width'/'height'/'random_state'"
                )

            try:
                width = kwargs.pop("width")
            except KeyError:
                width = None

            try:
                height = kwargs.pop("height")
            except KeyError:
                height = None

            try:
                random_state = kwargs.pop("random_state")
            except KeyError:
                random_state = None

            kwargs["position"] = position_jitter(
                width=width, height=height, random_state=random_state
            )
        geom_point.__init__(self, mapping, data, **kwargs)



import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineError
from ..mapping.aes import has_groups
from ..mapping.evaluation import after_stat
from ..utils import array_kind, jitter, resolution
from .binning import breaks_from_bins, breaks_from_binwidth
from .stat import stat
from .stat_density import compute_density


@document
class stat_sina(stat):
    """
    Compute Sina plot values

    {usage}

    Parameters
    ----------
    {common_parameters}
    binwidth : float
        The width of the bins. The default is to use bins that
        cover the range of the data. You should always override this
        value, exploring multiple widths to find the best to
        illustrate the stories in your data.
    bins : int (default: 50)
        Number of bins. Overridden by binwidth.
    method : 'density' or 'counts'
        Choose the method to spread the samples within the same bin
        along the x-axis. Available methods: "density", "counts"
        (can be abbreviated, e.g. "d"). See Details.
    maxwidth : float
        Control the maximum width the points can spread into.
        Values should be in the range (0, 1).
    adjust : float, optional (default: 1)
        Adjusts the bandwidth of the density kernel when
        ``method='density'`` (see density).
    bw : str or float, optional (default: 'nrd0')
        The bandwidth to use, If a float is given, it is the bandwidth.
        The :py:`str` choices are::

            'nrd0'
            'normal_reference'
            'scott'
            'silverman'

        ``nrd0`` is a port of ``stats::bw.nrd0`` in R; it is eqiuvalent
        to ``silverman`` when there is more than 1 value in a group.
    bin_limit : int (default: 1)
        If the samples within the same y-axis bin are more
        than `bin_limit`, the samples's X coordinates will be adjusted.
        This parameter is effective only when :py:`method='counts'`
    random_state : int or ~numpy.random.RandomState, optional
        Seed or Random number generator to use. If ``None``, then
        numpy global generator :class:`numpy.random` is used.
    scale : str (default: area)
        How to scale the sina groups. The options are::

            'area'   # Scale by the largest density/bin amoung the different
                     # sinas

            'count'  # areas are scaled proportionally to the number of points

            'width'  # Only scale according to the maxwidth parameter.

    See Also
    --------
    plotnine.geoms.geom_sina
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

         'quantile'  # quantile
         'group'     # group identifier

    Calculated aesthetics are accessed using the `after_stat` function.
    e.g. :py:`after_stat('quantile')`.
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_AES = {"xend": after_stat("scaled")}
    DEFAULT_PARAMS = {
        "geom": "sina",
        "position": "dodge",
        "na_rm": False,
        "binwidth": None,
        "bins": None,
        "method": "density",
        "bw": "nrd0",
        "maxwidth": None,
        "adjust": 1,
        "bin_limit": 1,
        "random_state": None,
        "scale": "area",
    }
    CREATES = {"scaled"}

    def setup_data(self, data):
        if (
            array_kind.continuous(data["x"])
            and not has_groups(data)
            and (data["x"] != data.loc["x", 0]).any()
        ):
            raise TypeError(
                "Continuous x aesthetic -- did you forget " "aes(group=...)?"
            )
        return data

    def setup_params(self, data):
        params = self.params.copy()
        random_state = params["random_state"]

        if params["maxwidth"] is None:
            params["maxwidth"] = resolution(data["x"], False) * 0.9

        if params["binwidth"] is None and self.params["bins"] is None:
            params["bins"] = 50

        if random_state is None:
            params["random_state"] = np.random
        elif isinstance(random_state, int):
            params["random_state"] = np.random.RandomState(random_state)

        # Required by compute_density
        params["kernel"] = "gau"  # It has to be a gaussian kernel
        params["cut"] = 0
        params["gridsize"] = None
        params["clip"] = (-np.inf, np.inf)
        params["n"] = 512
        return params

    @classmethod
    def compute_panel(cls, data, scales, **params):
        maxwidth = params["maxwidth"]
        random_state = params["random_state"]
        fuzz = 1e-8
        y_dim = scales.y.dimension()
        y_dim_fuzzed = (y_dim[0] - fuzz, y_dim[1] + fuzz)

        if params["binwidth"] is not None:
            params["bins"] = breaks_from_binwidth(
                y_dim_fuzzed, params["binwidth"]
            )
        else:
            params["bins"] = breaks_from_bins(y_dim_fuzzed, params["bins"])

        data = super(cls, stat_sina).compute_panel(data, scales, **params)

        if not len(data):
            return data

        if params["scale"] == "area":
            data["sinawidth"] = data["density"] / data["density"].max()
        elif params["scale"] == "count":
            data["sinawidth"] = (
                data["density"]
                / data["density"].max()
                * data["n"]
                / data["n"].max()
            )
        elif params["scale"] == "width":
            data["sinawidth"] = data["scaled"]
        else:
            msg = "Unknown scale value '{}'"
            raise PlotnineError(msg.format(params["scale"]))

        data["xmin"] = data["x"] - maxwidth / 2
        data["xmax"] = data["x"] + maxwidth / 2
        data["x_diff"] = (
            random_state.uniform(-1, 1, len(data))
            * maxwidth
            * data["sinawidth"]
            / 2
        )
        data["width"] = maxwidth

        # jitter y values if the input is input is integer
        if (data["y"] == np.floor(data["y"])).all():
            data["y"] = jitter(data["y"], random_state=random_state)

        return data

    @classmethod
    def compute_group(cls, data, scales, **params):
        maxwidth = params["maxwidth"]
        bins = params["bins"]
        bin_limit = params["bin_limit"]
        weight = None

        if len(data) == 0:
            return pd.DataFrame()

        elif len(data) < 3:
            data["density"] = 0
            data["scaled"] = 1
        elif params["method"] == "density":
            from scipy.interpolate import interp1d

            # density kernel estimation
            range_y = data["y"].min(), data["y"].max()
            dens = compute_density(data["y"], weight, range_y, **params)
            densf = interp1d(
                dens["x"],
                dens["density"],
                bounds_error=False,
                fill_value="extrapolate",  # pyright: ignore
            )
            data["density"] = densf(data["y"])
            data["scaled"] = data["density"] / dens["density"].max()
        else:
            # bin based estimation
            bin_index = pd.cut(
                data["y"], bins, include_lowest=True, labels=False
            )
            data["density"] = (
                pd.Series(bin_index)
                .groupby(bin_index)
                .apply(len)[bin_index]
                .values
            )
            data.loc[data["density"] <= bin_limit, "density"] = 0
            data["scaled"] = data["density"] / data["density"].max()

        # Compute width if x has multiple values
        if len(data["x"].unique()) > 1:
            width = np.ptp(data["x"]) * maxwidth
        else:
            width = maxwidth

        data["width"] = width
        data["n"] = len(data)
        data["x"] = np.mean([data["x"].max(), data["x"].min()])

        return data

    def finish_layer(self, data, params):
        # Rescale x in case positions have been adjusted
        x_mod = (data["xmax"] - data["xmin"]) / data["width"]
        data["x"] = data["x"] + data["x_diff"] * x_mod
        return data



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from .stat import stat

if typing.TYPE_CHECKING:
    from typing import Any, Optional

    from plotnine.typing import FloatArrayLike


@document
class stat_ellipse(stat):
    """
    Calculate normal confidence interval ellipse

    {usage}

    Parameters
    ----------
    {common_parameters}
    type : str in ['t', 'norm', 'euclid'] (default: 't')
        The type of ellipse.

        - ``'t'`` - assumes a multivariate t-distribution, and
        - ``'norm'`` - assumes a multivariate normal distribution.
        - ``'euclid'`` - draws a circle with the radius equal to
          `level`, representing the euclidean distance from the
          center.

    level : float, optional (default: 0.95)
        The confidence level at which to draw the ellipse.
    segments : int, optional (default: 51)
        Number of segments to be used in drawing the ellipse.
    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "geom": "path",
        "position": "identity",
        "na_rm": False,
        "type": "t",
        "level": 0.95,
        "segments": 51,
    }

    @classmethod
    def compute_group(cls, data, scales, **params):
        import scipy.stats as stats
        from scipy import linalg

        level = params["level"]
        segments = params["segments"]
        type_ = params["type"]

        dfn = 2
        dfd = len(data) - 1

        if dfd < 3:
            warn("Too few points to calculate an ellipse", PlotnineWarning)
            return pd.DataFrame({"x": [], "y": []})

        m = np.asarray(data[["x", "y"]])

        # The stats used to create the ellipse
        if type_ == "t":
            res = cov_trob(m)
            cov = res["cov"]
            center = res["center"]
        elif type_ == "norm":
            cov = np.cov(m, rowvar=False)
            center = np.mean(m, axis=0)
        elif type_ == "euclid":
            cov = np.cov(m, rowvar=False)
            cov = np.diag(np.repeat(np.diag(cov).min(), 2))
            center = np.mean(m, axis=0)
        else:
            raise ValueError(f"Unknown value for type={type_}")

        # numpy's cholesky function does not gaurantee upper/lower
        # triangular factorization.
        chol_decomp = linalg.cholesky(cov, lower=False)

        # Parameters of the ellipse
        if type_ == "euclid":
            radius = level / chol_decomp.max()
        else:
            radius = np.sqrt(dfn * stats.f.ppf(level, dfn, dfd))

        space = np.linspace(0, 2 * np.pi, segments)

        # Catesian coordinates
        unit_circle = np.column_stack([np.cos(space), np.sin(space)])
        res = center + radius * np.dot(unit_circle, chol_decomp)

        return pd.DataFrame({"x": res[:, 0], "y": res[:, 1]})


def cov_trob(
    x,
    wt: Optional[FloatArrayLike] = None,
    cor=False,
    center: FloatArrayLike | bool = True,
    nu=5,
    maxit=25,
    tol=0.01,
):
    """
    Covariance Estimation for Multivariate t Distribution

    Estimates a covariance or correlation matrix assuming the
    data came from a multivariate t distribution: this provides
    some degree of robustness to outlier without giving a high
    breakdown point.

    **credit**: This function a port of the R function
    ``MASS::cov.trob``.

    Parameters
    ----------
    x : array
        data matrix. Missing values (NaNs) are not allowed.
    wt : array
        A vector of weights for each case: these are treated as
        if the case i actually occurred ``wt[i]`` times.
    cor : bool
        Flag to choose between returning the correlation
        (``cor=True``) or covariance (``cor=False``) matrix.
    center : array or bool
        A logical value or a numeric vector providing the location
        about which the covariance is to be taken.
        If ``center=False``, no centering is done; if
        ``center=True`` the MLE of the location vector is used.
    nu : int
        'degrees of freedom' for the multivariate t distribution.
        Must exceed 2 (so that the covariance matrix is finite).
    maxit : int
        Maximum number of iterations in fitting.
    tol : float
        Convergence tolerance for fitting.

    Returns
    -------
    out : dict
        A dictionary with with the following key-value

        - ``cov`` : the fitted covariance matrix.
        - ``center`` : the estimated or specified location vector.
        - ``wt`` : the specified weights: only returned if the
           wt argument was given.
        - ``n_obs`` : the number of cases used in the fitting.
        - ``cor`` : the fitted correlation matrix: only returned
          if ``cor=True``.
        - ``call`` : The matched call.
        - ``iter`` : The number of iterations used.

    References
    ----------
    - J. T. Kent, D. E. Tyler and Y. Vardi (1994) A curious likelihood
      identity for the multivariate t-distribution. *Communications in
      Statistics-Simulation and Computation* **23**, 441-453.

    - Venables, W. N. and Ripley, B. D. (1999) *Modern Applied
      Statistics with S-PLUS*. Third Edition. Springer.

    """
    from scipy import linalg

    def test_values(x):
        if pd.isnull(x).any() or np.isinf(x).any():
            raise ValueError("Missing or infinite values in 'x'")

    def scale_simp(x, center, n, p):
        return x - np.repeat([center], n, axis=0)

    x = np.asarray(x)
    n, p = x.shape
    test_values(x)
    ans: dict[str, Any] = {}

    # wt
    if wt is None:
        wt = np.ones(n)
    else:
        wt = np.asarray(wt)
        ans["wt0"] = wt

        if len(wt) != n:
            raise ValueError(
                "length of 'wt' must equal number of observations."
            )
        if any(wt < 0):
            raise ValueError("Negative weights not allowed.")
        if not np.sum(wt):
            raise ValueError("No positive weights.")

        x = x[wt > 0, :]
        wt = wt[wt > 0]
        n, _ = x.shape

    wt = wt[:, np.newaxis]

    # loc
    use_loc = False
    if isinstance(center, bool):
        if center:
            loc = np.sum(wt * x, axis=0) / wt.sum()
            use_loc = True
        else:
            loc = np.zeros(p)
    else:
        if len(center) != p:
            raise ValueError("'center' is not the right length")
        loc = p

    w = wt * (1 + p / nu)
    for iteration in range(maxit):
        w0 = w
        X = scale_simp(x, loc, n, p)
        _, s, v = linalg.svd(np.sqrt(w / np.sum(w)) * X)
        # wX = X @ v.T @ np.diag(np.full(p, 1/s))
        wX = np.dot(np.dot(X, v.T), np.diag(np.full(p, 1 / s)))
        # Q = np.squeeze((wX**2) @ np.ones(p))
        Q = np.squeeze(np.dot(wX**2, np.ones(p)))
        w = (wt * (nu + p)) / (nu + Q)[:, np.newaxis]
        if use_loc:
            loc = np.sum(w * x, axis=0) / w.sum()
        if all(np.abs(w - w0) < tol):
            break
    else:  # nobreak
        _c1 = np.mean(w) - np.mean(wt) > tol
        _c2 = np.abs(np.mean(w * Q) / p - 1) > tol  # pyright: ignore
        if _c1 and _c2:
            warn("Convergence probably failed.", PlotnineWarning)

    _a = np.sqrt(w) * X  # pyright: ignore[reportUnboundVariable]
    # cov = (_a.T @ _a) / np.sum(wt)
    cov = np.dot(_a.T, _a) / np.sum(wt)

    if cor:
        sd = np.sqrt(np.diag(cov))
        ans["cor"] = (cov / sd) / np.repeat([sd], p, axis=0).T

    ans.update(
        cov=cov,
        center=loc,
        n_obs=n,
        iter=iteration,  # pyright: ignore[reportUnboundVariable]
    )
    return ans
